# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import pandas as pd
import warnings

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def required_std_type_parameters(element="line"):
    if element == "line":
        required = ["c_nf_per_km", "r_ohm_per_km", "x_ohm_per_km", "max_i_ka"]
    elif element == "line_dc":
        required = ["r_ohm_per_km","max_i_ka"]
    elif element == "trafo":
        required = ["sn_mva", "vn_hv_kv", "vn_lv_kv", "vk_percent", "vkr_percent",
                    "pfe_kw", "i0_percent", "shift_degree"]
    elif element == "trafo3w":
        required = ["sn_hv_mva", "sn_mv_mva", "sn_lv_mva", "vn_hv_kv", "vn_mv_kv", "vn_lv_kv",
                    "vk_hv_percent", "vk_mv_percent", "vk_lv_percent", "vkr_hv_percent",
                    "vkr_mv_percent", "vkr_lv_percent", "pfe_kw", "i0_percent", "shift_mv_degree",
                    "shift_lv_degree"]
    elif element == "fuse":
        required = ["fuse_type", "i_rated_a"]
    else:
        raise ValueError("Unknown element type %s" % element)
    return required


def create_std_type(net, data, name, element="line", overwrite=True, check_required=True):
    """
    Creates type data in the type database. The parameters that are used for
    the loadflow have to be at least contained in data. These parameters are:

        - c_nf_per_km, r_ohm_per_km, x_ohm_per_km and max_i_ka (for lines)
        - sn_mva, vn_hv_kv, vn_lv_kv, vk_percent, vkr_percent, pfe_kw, i0_percent, shift_degree* (for transformers)
        - sn_hv_mva, sn_mv_mva, sn_lv_mva, vn_hv_kv, vn_mv_kv, vn_lv_kv, vk_hv_percent, vk_mv_percent, vk_lv_percent,
            vkr_hv_percent, vkr_mv_percent, vkr_lv_percent, pfe_kw, i0_percent, shift_mv_degree*, shift_lv_degree* (for 3-winding-transformers)

    additional parameters can be added and later loaded into pandapower with the function
    "parameter_from_std_type".

    ** only considered in loadflow if calculate_voltage_angles = True

    The standard type is saved into the pandapower library of the given network by default.

    INPUT:
        **net** - The pandapower network

        **data** - dictionary of standard type parameters

        **name** - name of the standard type as string

        **element** - "line", "trafo" or "trafo3w"

    EXAMPLE:

    >>> line_data = {"c_nf_per_km": 0, "r_ohm_per_km": 0.642, "x_ohm_per_km": 0.083, "max_i_ka": 0.142, "type": "cs", "q_mm2": 50, "alpha": 4.03e-3}
    >>> pandapower.create_std_type(net, line_data, "NAYY 4×50 SE", element='line')
    >>> # Three phase line creation:
    >>> pandapower.create_std_type(net, {"r_ohm_per_km": 0.1941, "x_ohm_per_km": 0.07476991,
                    "c_nf_per_km": 1160., "max_i_ka": 0.421,
                    "endtemp_degree": 70.0, "r0_ohm_per_km": 0.7766,
                    "x0_ohm_per_km": 0.2990796,
                    "c0_nf_per_km":  496.2}, name="unsymmetric_line_type",element = "line")
    >>> # Three phase transformer creation
    >>> pp.create_std_type(net, {"sn_mva": 1.6,
            "vn_hv_kv": 10,
            "vn_lv_kv": 0.4,
            "vk_percent": 6,
            "vkr_percent": 0.78125,
            "pfe_kw": 2.7,
            "i0_percent": 0.16875,
            "shift_degree": 0,
            "vector_group": vector_group,
            "tap_side": "lv",
            "tap_neutral": 0,
            "tap_min": -2,
            "tap_max": 2,
            "tap_step_degree": 0,
            "tap_step_percent": 2.5,
            "tap_changer_type": "Ratio",
            "vk0_percent": 6,
            "vkr0_percent": 0.78125,
            "mag0_percent": 100,
            "mag0_rx": 0.,
            "si0_hv_partial": 0.9,}, name='Unsymmetric_trafo_type', element="trafo")
    """

    if type(data) != dict:
        raise UserWarning("type data has to be given as a dictionary of parameters")

    if check_required:
        missing = [par for par in required_std_type_parameters(element) if par not in data]
        if len(missing):
            raise UserWarning("%s are required as %s type parameters." % (missing, element))
    library = net.std_types[element]
    if overwrite or not (name in library):
        library.update({name: data})


def create_std_types(net, data, element="line", overwrite=True, check_required=True):
    """
    Creates multiple standard types in the type database.

    INPUT:
        **net** - The pandapower network
        **data** - dictionary of standard type parameter sets

        **element** - "line", "line_dc", "trafo" or "trafo3w"

    EXAMPLE:

    >>> linetypes = {"typ1": {"r_ohm_per_km": 0.01, "x_ohm_per_km": 0.02, "c_nf_per_km": 10, "max_i_ka": 0.4, "type": "cs"},
    >>>              "typ2": {"r_ohm_per_km": 0.015, "x_ohm_per_km": 0.01, "c_nf_per_km": 30, "max_i_ka": 0.3, "type": "cs"}}
    >>> pp.create_std_types(net, data=linetypes, element="line")

    """
    for name, typdata in data.items():
        create_std_type(net, data=typdata, name=name, element=element, overwrite=overwrite,
                        check_required=check_required)


def copy_std_types(to_net, from_net, element="line", overwrite=True):
    """
    Transfers all standard types of one network to another.

    INPUT:

        **to_net** - The pandapower network to which the standard types are copied

        **from_net** - The pandapower network from which the standard types are taken

        **element** - "line" or "trafo"

        **overwrite** - if True, overwrites standard types which already exist in to_net

    """
    for name, typdata in from_net.std_types[element].items():
        create_std_type(to_net, typdata, name, element=element, overwrite=overwrite)


def load_std_type(net, name, element="line"):
    """
    Loads standard type data from the linetypes database. Issues a warning if
    linetype is unknown.

    INPUT:
        **net** - The pandapower network

        **name** - name of the standard type as string

        **element** -  "line", "line_dc", "trafo" or "trafo3w"

    OUTPUT:
        **typedata** - dictionary containing type data
    """
    library = net.std_types[element]
    if name in library:
        return library[name]
    else:
        raise UserWarning("Unknown standard %s type %s" % (element, name))


def std_type_exists(net, name, element="line"):
    """
    Checks if a standard type exists.

    INPUT:
        **net** - pandapower Network

        **name** - name of the standard type as string

        **element** - type of element ("line" or "trafo")

    OUTPUT:
        **exists** - True if standard type exists, False otherwise
    """
    library = net.std_types[element]
    return name in library


def delete_std_type(net, name, element="line"):
    """
    Deletes standard type parameters from database.

    INPUT:
        **net** - pandapower Network

        **name** - name of the standard type as string

        **element** - type of element ("line" or "trafo")

    """
    library = net.std_types[element]
    if name in library:
        del library[name]
    else:
        raise UserWarning("Unknown standard %s type %s" % (element, name))


def rename_std_type(net, old_name, new_name, element="line"):
    """Renames an existing standard type in the standard type library and the element table.

    Parameters
    ----------
    net : pp.pandapowerNet
        pandapower net
    old_name : str
        old name to be replaced
    new_name : str
        new name of the standard type
    element : str, optional
        type of element, by default "line"
    """
    library = net.std_types[element]
    if old_name not in library:
        raise UserWarning(f"Unknown {element} standard type '{old_name}'.")
    if new_name in library:
        raise UserWarning(f"{element} standard type '{new_name}' already exists.")
    library[new_name] = library.pop(old_name)
    net[element].loc[net[element].std_type == old_name, "std_type"] = new_name


def available_std_types(net, element="line"):
    """
    Returns all standard types available for this network as a table.

    INPUT:
        **net** - pandapower Network

        **element** - type of element ("line" or "trafo")

    OUTPUT:
        **typedata** - table of standard type parameters

    """
    std_types = pd.DataFrame(net.std_types[element]).T
    try:
        return std_types.infer_objects()
    except AttributeError:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return std_types.convert_objects()


def parameter_from_std_type(net, parameter, element="line", fill=None):
    """
    Loads standard types data for a parameter, which can be used to add an additional parameter,
    that is not included in the original pandapower datastructure but is available in the standard
    type database.

    INPUT:
        **net** - pandapower network

        **parameter** - name of parameter as string

        **element** - type of element ("line" or "trafo")

        **fill** - fill-value that is assigned to all lines/trafos without
            a value for the parameter, either because the line/trafo has no type or because the
            type does not have a value for the parameter

    EXAMPLE:

        >>> from pandapower import parameter_from_std_type
        >>> from pandapower.networks import simple_mv_open_ring_net
        >>>
        >>> net = simple_mv_open_ring_net()
        >>> parameter_from_std_type(net, "q_mm2")
    """
    if parameter not in net[element]:
        net[element][parameter] = fill
    for typ in net[element].std_type.unique():
        if pd.isnull(typ) or not std_type_exists(net, typ, element):
            continue
        typedata = load_std_type(net, name=typ, element=element)
        if parameter in typedata:
            util = net[element].loc[net[element].std_type == typ].index
            net[element].loc[util, parameter] = typedata[parameter]
    if fill is not None:
        net[element].loc[pd.isnull(net[element][parameter]).values, parameter] = fill


def change_std_type(net, eid, name, element="line"):
    """
    Changes the type of a given element in pandapower. Changes only parameter that are given
    for the type.

    INPUT:
        **net** - pandapower network

        **eid** - element index (either line or transformer index)

        **element** - type of element ("line" or "trafo")

        **name** - name of the new standard type

    """
    type_param = load_std_type(net, name, element)
    table = net[element]
    for column in table.columns:
        if column in type_param:
            table.at[eid, column] = type_param[column]
    table.at[eid, "std_type"] = name


def find_std_type_by_parameter(net, data, element="line", epsilon=0.):
    """
    Searches for a std_type that fits all values given in the data dictionary with the margin of
    epsilon.

    INPUT:
        **net** - pandapower network

        **data** - dictionary of standard type parameters

        **element** - type of element ("line" or "trafo")

        **epsilon** - tolerance margin for parameter comparison

    OUTPUT:
        **fitting_types** - list of fitting types or empty list
    """
    fitting_types = []
    assert epsilon >= 0
    for name, stp in net.std_types[element].items():
        for p, v in list(data.items()):
            if isinstance(v, float):
                if abs(v - stp[p]) > epsilon:
                    break
            elif stp[p] != v:
                break
        else:
            fitting_types.append(name)
    return fitting_types


def find_std_type_alternative(net, data, element = "line", voltage_rating = "", epsilon = 0.):
    """
        Searches for a std_type that fits all values given in the standard types library with the margin of
        epsilon.

        INPUT:
            **net** - pandapower network

            **data** - dictionary of standard type parameters

            **element** - type of element ("line" or "trafo")

            **voltage_rating** - voltage rating of the cable ("HV" or "MV" or "LV")

            **epsilon** - tolerance margin for parameter comparison

        OUTPUT:
            **fitting_types** - list of fitting types or empty list
        """

    assert epsilon >= 0
    linetypes = basic_line_std_types()
    possible_alternatives = []
    fitting_types = []
    for p, v in linetypes.items():
        if voltage_rating == v.get("voltage_rating"):
            possible_alternatives.append((p, v))
    for name, stp in possible_alternatives:
        for p, v in list(data.items()):
            if isinstance(v, float):
                if abs(v - stp[p]) > epsilon:
                    break
            elif stp[p] != v:
                break
        else:
            fitting_types.append(name)
    return fitting_types


def add_zero_impedance_parameters(net):
    """
    Adds all parameters required for zero sequence impedance calculations.

    INPUT:
        **net** - pandapower network

        zero sequence parameters of lines and transformers in pandapower networks
        are entered using std_type.

        This function adds them to the pandas dataframe


    OUTPUT:
        Now, net has all the zero sequence  parameters
    """
    parameter_from_std_type(net, "vector_group", element="trafo")
    parameter_from_std_type(net, "vk0_percent", element="trafo")
    parameter_from_std_type(net, "vkr0_percent", element="trafo")
    parameter_from_std_type(net, "mag0_percent", element="trafo")
    parameter_from_std_type(net, "mag0_rx", element="trafo")
    parameter_from_std_type(net, "si0_hv_partial", element="trafo")
    parameter_from_std_type(net, "g0_nf_per_km")
    parameter_from_std_type(net, "c0_nf_per_km")
    parameter_from_std_type(net, "r0_ohm_per_km")
    parameter_from_std_type(net, "x0_ohm_per_km")
    parameter_from_std_type(net, "endtemp_degree")

    # add zero seq. parameters for ext_grid and apply standard values
    if 's_sc_max_mva' not in net.ext_grid.columns:
        net.ext_grid['s_sc_max_mva'] = 1000
    if 'rx_max' not in net.ext_grid.columns:
        net.ext_grid['rx_max'] = 0.1
    if 'x0x_max' not in net.ext_grid.columns:
        net.ext_grid['x0x_max'] = 1
    if 'r0x0_max' not in net.ext_grid.columns:
        net.ext_grid['r0x0_max'] = 0.1


def add_temperature_coefficient(net, fill=None):
    """
    Adds alpha parameter for calculations of line temperature
    Args:
        fill: fill value for when the parameter in std_type is missing, e.g. 4.03e-3 for aluminum
                or  3.93e-3 for copper

    """
    parameter_from_std_type(net, "alpha", fill=fill)
    parameter_from_std_type(net, "alpha", fill=fill,element="line_dc")


def basic_line_std_types():

    alpha_al = 4.03e-3
    alpha_cu = 3.93e-3

    linetypes = {
        # Cables, all from S.744, Heuck: Elektrische Energieversorgung - Vierweg+Teubner 2013
        # additional MV cables from Werth: Netzberechnung mit Erzeugungsporfilen (Dreiecksverlegung)
        # Low Voltage
        "NAYY 4x50 SE":
        {"c_nf_per_km": 210,
            "r_ohm_per_km": 0.642,
            "x_ohm_per_km": 0.083,
            "max_i_ka": 0.142,
            "type": "cs",
            "q_mm2": 50,
            "alpha": alpha_al,
            "voltage_rating": "LV"},
        "NAYY 4x120 SE":
        {"c_nf_per_km": 264,
            "r_ohm_per_km": 0.225,
            "x_ohm_per_km": 0.080,
            "max_i_ka": 0.242,
            "type": "cs",
            "q_mm2": 120,
            "alpha": alpha_al,
         "voltage_rating": "LV"},
        "NAYY 4x150 SE":
        {"c_nf_per_km": 261,
            "r_ohm_per_km": 0.208,
            "x_ohm_per_km": 0.080,
            "max_i_ka": 0.270,
            "type": "cs",
            "q_mm2": 150,
            "alpha": alpha_al,
         "voltage_rating": "LV"},

        # Medium Voltage
        "NA2XS2Y 1x95 RM/25 12/20 kV":
        {"c_nf_per_km": 216,
            "r_ohm_per_km": 0.313,
            "x_ohm_per_km": 0.132,
            "max_i_ka": 0.252,
            "type": "cs",
            "q_mm2": 95,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "NA2XS2Y 1x185 RM/25 12/20 kV":
        {"c_nf_per_km": 273,
            "r_ohm_per_km": 0.161,
            "x_ohm_per_km": 0.117,
            "max_i_ka": 0.362,
            "type": "cs",
            "q_mm2": 185,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "NA2XS2Y 1x240 RM/25 12/20 kV":
        {"c_nf_per_km": 304,
            "r_ohm_per_km": 0.122,
            "x_ohm_per_km": 0.112,
            "max_i_ka": 0.421,
            "type": "cs",
            "q_mm2": 240,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "NA2XS2Y 1x95 RM/25 6/10 kV":
        {"c_nf_per_km": 315,
            "r_ohm_per_km": 0.313,
            "x_ohm_per_km": 0.123,
            "max_i_ka": 0.249,
            "type": "cs",
            "q_mm2": 95,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "NA2XS2Y 1x185 RM/25 6/10 kV":
        {"c_nf_per_km": 406,
            "r_ohm_per_km": 0.161,
            "x_ohm_per_km": 0.110,
            "max_i_ka": 0.358,
            "type": "cs",
            "q_mm2": 185,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "NA2XS2Y 1x240 RM/25 6/10 kV":
        {"c_nf_per_km": 456,
            "r_ohm_per_km": 0.122,
            "x_ohm_per_km": 0.105,
            "max_i_ka": 0.416,
            "type": "cs",
            "q_mm2": 240,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        # additional MV cables
        "NA2XS2Y 1x150 RM/25 12/20 kV":
        {"c_nf_per_km": 250,
            "r_ohm_per_km": 0.206,
            "x_ohm_per_km": 0.116,
            "max_i_ka": 0.319,
            "type": "cs",
            "q_mm2": 150,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "NA2XS2Y 1x120 RM/25 12/20 kV":
        {"c_nf_per_km": 230,
            "r_ohm_per_km": 0.253,
            "x_ohm_per_km": 0.119,
            "max_i_ka": 0.283,
            "type": "cs",
            "q_mm2": 120,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "NA2XS2Y 1x70 RM/25 12/20 kV":
        {"c_nf_per_km": 190,
            "r_ohm_per_km": 0.443,
            "x_ohm_per_km": 0.132,
            "max_i_ka": 0.220,
            "type": "cs",
            "q_mm2": 70,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "NA2XS2Y 1x150 RM/25 6/10 kV":
        {"c_nf_per_km": 360,
            "r_ohm_per_km": 0.206,
            "x_ohm_per_km": 0.110,
            "max_i_ka": 0.315,
            "type": "cs",
            "q_mm2": 150,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "NA2XS2Y 1x120 RM/25 6/10 kV":
        {"c_nf_per_km": 340,
            "r_ohm_per_km": 0.253,
            "x_ohm_per_km": 0.113,
            "max_i_ka": 0.280,
            "type": "cs",
            "q_mm2": 120,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "NA2XS2Y 1x70 RM/25 6/10 kV":
        {"c_nf_per_km": 280,
            "r_ohm_per_km": 0.443,
            "x_ohm_per_km": 0.123,
            "max_i_ka": 0.217,
            "type": "cs",
            "q_mm2": 70,
            "alpha": alpha_al,
         "voltage_rating": "MV"},

        # High Voltage
        "N2XS(FL)2Y 1x120 RM/35 64/110 kV":
        {"c_nf_per_km": 112,
            "r_ohm_per_km": 0.153,
            "x_ohm_per_km": 0.166,
            "max_i_ka": 0.366,
            "type": "cs",
            "q_mm2": 120,
            "alpha": alpha_cu,
         "voltage_rating": "HV"},
        "N2XS(FL)2Y 1x185 RM/35 64/110 kV":
        {"c_nf_per_km": 125,
            "r_ohm_per_km": 0.099,
            "x_ohm_per_km": 0.156,
            "max_i_ka": 0.457,
            "type": "cs",
            "q_mm2": 185,
            "alpha": alpha_cu,
         "voltage_rating": "HV"},
        "N2XS(FL)2Y 1x240 RM/35 64/110 kV":
        {"c_nf_per_km": 135,
            "r_ohm_per_km": 0.075,
            "x_ohm_per_km": 0.149,
            "max_i_ka": 0.526,
            "type": "cs",
            "q_mm2": 240,
            "alpha": alpha_cu,
         "voltage_rating": "HV"},
        "N2XS(FL)2Y 1x300 RM/35 64/110 kV":
        {"c_nf_per_km": 144,
            "r_ohm_per_km": 0.060,
            "x_ohm_per_km": 0.144,
            "max_i_ka": 0.588,
            "type": "cs",
            "q_mm2": 300,
            "alpha": alpha_cu,
         "voltage_rating": "HV"},

        # Overhead Lines, all from S.742f, Heuck: Elektrische Energieversorgung -
        # Vierweg+Teubner 2013
        # 679/86 110 from S. 362, Flosdorff, Hilgarth: Elektrische Energieverteilung - Teubner 2005

        # Low Voltage
        "15-AL1/3-ST1A 0.4":
        {"c_nf_per_km": 11,
            "r_ohm_per_km": 1.8769,
            "x_ohm_per_km": 0.35,
            "max_i_ka": 0.105,
            "type": "ol",
            "q_mm2": 16,
            "alpha": alpha_al,
         "voltage_rating": "LV"},
        "24-AL1/4-ST1A 0.4":
        {"c_nf_per_km": 11.25,
            "r_ohm_per_km": 1.2012,
            "x_ohm_per_km": 0.335,
            "max_i_ka": 0.140,
            "type": "ol",
            "q_mm2": 24,
            "alpha": alpha_al,
         "voltage_rating": "LV"},
        "48-AL1/8-ST1A 0.4":
        {"c_nf_per_km": 12.2,
            "r_ohm_per_km": 0.5939,
            "x_ohm_per_km": 0.3,
            "max_i_ka": .210,
            "type": "ol",
            "q_mm2": 48,
            "alpha": alpha_al,
         "voltage_rating": "LV"},
        "94-AL1/15-ST1A 0.4":
        {"c_nf_per_km": 13.2,
            "r_ohm_per_km": 0.3060,
            "x_ohm_per_km": 0.29,
            "max_i_ka": 0.350,
            "type": "ol",
            "q_mm2": 94,
            "alpha": alpha_al,
         "voltage_rating": "LV"},

        # Medium Voltage
        "34-AL1/6-ST1A 10.0":
        {"c_nf_per_km": 9.7,
            "r_ohm_per_km": 0.8342,
            "x_ohm_per_km": 0.36,
            "max_i_ka": 0.170,
            "type": "ol",
            "q_mm2": 34,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "48-AL1/8-ST1A 10.0":
        {"c_nf_per_km": 10.1,
            "r_ohm_per_km": 0.5939,
            "x_ohm_per_km": 0.35,
            "max_i_ka": 0.210,
            "type": "ol",
            "q_mm2": 48,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "70-AL1/11-ST1A 10.0":
        {"c_nf_per_km": 10.4,
            "r_ohm_per_km": 0.4132,
            "x_ohm_per_km": 0.339,
            "max_i_ka": 0.290,
            "type": "ol",
            "q_mm2": 70,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "94-AL1/15-ST1A 10.0":
        {"c_nf_per_km": 10.75,
            "r_ohm_per_km": 0.3060,
            "x_ohm_per_km": 0.33,
            "max_i_ka": 0.350,
            "type": "ol",
            "q_mm2": 94,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "122-AL1/20-ST1A 10.0":
        {"c_nf_per_km": 11.1,
            "r_ohm_per_km": 0.2376,
            "x_ohm_per_km": 0.323,
            "max_i_ka": 0.410,
            "type": "ol",
            "q_mm2": 122,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "149-AL1/24-ST1A 10.0":
        {"c_nf_per_km": 11.25,
            "r_ohm_per_km": 0.1940,
            "x_ohm_per_km": 0.315,
            "max_i_ka": 0.470,
            "type": "ol",
            "q_mm2": 149,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "34-AL1/6-ST1A 20.0":
        {"c_nf_per_km": 9.15,
            "r_ohm_per_km": 0.8342,
            "x_ohm_per_km": 0.382,
            "max_i_ka": 0.170,
            "type": "ol",
            "q_mm2": 34,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "48-AL1/8-ST1A 20.0":
        {"c_nf_per_km": 9.5,
            "r_ohm_per_km": 0.5939,
            "x_ohm_per_km": 0.372,
            "max_i_ka": 0.210,
            "type": "ol",
            "q_mm2": 48,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "70-AL1/11-ST1A 20.0":
        {"c_nf_per_km": 9.7,
            "r_ohm_per_km": 0.4132,
            "x_ohm_per_km": 0.36,
            "max_i_ka": 0.290,
            "type": "ol",
            "q_mm2": 70,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "94-AL1/15-ST1A 20.0":
        {"c_nf_per_km": 10,
            "r_ohm_per_km": 0.3060,
            "x_ohm_per_km": 0.35,
            "max_i_ka": 0.350,
            "type": "ol",
            "q_mm2": 94,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "122-AL1/20-ST1A 20.0":
        {"c_nf_per_km": 10.3,
            "r_ohm_per_km": 0.2376,
            "x_ohm_per_km": 0.344,
            "max_i_ka": 0.410,
            "type": "ol",
            "q_mm2": 122,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "149-AL1/24-ST1A 20.0":
        {"c_nf_per_km": 10.5,
            "r_ohm_per_km": 0.1940,
            "x_ohm_per_km": 0.337,
            "max_i_ka": 0.470,
            "type": "ol",
            "q_mm2": 149,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "184-AL1/30-ST1A 20.0":
        {"c_nf_per_km": 10.75,
            "r_ohm_per_km": 0.1571,
            "x_ohm_per_km": 0.33,
            "max_i_ka": 0.535,
            "type": "ol",
            "q_mm2": 184,
            "alpha": alpha_al,
         "voltage_rating": "MV"},
        "243-AL1/39-ST1A 20.0":
        {"c_nf_per_km": 11,
            "r_ohm_per_km": 0.1188,
            "x_ohm_per_km": 0.32,
            "max_i_ka": 0.645,
            "type": "ol",
            "q_mm2": 243,
            "alpha": alpha_al,
         "voltage_rating": "MV"},

        # High Voltage
        # c acd x values are estimated for 4 m conductor distance, single bundle and "Donaumast"
        "48-AL1/8-ST1A 110.0":
        {"c_nf_per_km": 8,
            "r_ohm_per_km": 0.5939,
            "x_ohm_per_km": 0.46,
            "max_i_ka": 0.210,
            "type": "ol",
            "q_mm2": 48,
            "alpha": alpha_al,
         "voltage_rating": "HV"},
        "70-AL1/11-ST1A 110.0":
        {"c_nf_per_km": 8.4,
            "r_ohm_per_km": 0.4132,
            "x_ohm_per_km": 0.45,
            "max_i_ka": 0.290,
            "type": "ol",
            "q_mm2": 70,
            "alpha": alpha_al,
         "voltage_rating": "HV"},
        "94-AL1/15-ST1A 110.0":
        {"c_nf_per_km": 8.65,
            "r_ohm_per_km": 0.3060,
            "x_ohm_per_km": 0.44,
            "max_i_ka": 0.350,
            "type": "ol",
            "q_mm2": 94,
            "alpha": alpha_al,
         "voltage_rating": "HV"},
        "122-AL1/20-ST1A 110.0":
        {"c_nf_per_km": 8.5,
            "r_ohm_per_km": 0.2376,
            "x_ohm_per_km": 0.43,
            "max_i_ka": 0.410,
            "type": "ol",
            "q_mm2": 122,
            "alpha": alpha_al,
         "voltage_rating": "HV"},
        "149-AL1/24-ST1A 110.0":
        {"c_nf_per_km": 8.75,
            "r_ohm_per_km": 0.1940,
            "x_ohm_per_km": 0.41,
            "max_i_ka": 0.470,
            "type": "ol",
            "q_mm2": 149,
            "alpha": alpha_al,
         "voltage_rating": "HV"},
        "184-AL1/30-ST1A 110.0":
        {"c_nf_per_km": 8.8,
            "r_ohm_per_km": 0.1571,
            "x_ohm_per_km": 0.4,
            "max_i_ka": 0.535,
            "type": "ol",
            "q_mm2": 184,
            "alpha": alpha_al,
         "voltage_rating": "HV"},
        "243-AL1/39-ST1A 110.0":
        {"c_nf_per_km": 9,
            "r_ohm_per_km": 0.1188,
            "x_ohm_per_km": 0.39,
            "max_i_ka": 0.645,
            "type": "ol",
            "q_mm2": 243,
            "alpha": alpha_al,
         "voltage_rating": "HV"},
        "305-AL1/39-ST1A 110.0":
        {"c_nf_per_km": 9.2,
            "r_ohm_per_km": 0.0949,
            "x_ohm_per_km": 0.38,
            "max_i_ka": 0.74,
            "type": "ol",
            "q_mm2": 305,
            "alpha": alpha_al,
         "voltage_rating": "HV"},
        "490-AL1/64-ST1A 110.0":
        {"c_nf_per_km": 9.75,
            "r_ohm_per_km": 0.059,
            "x_ohm_per_km": 0.37,
            "max_i_ka": 0.960,
            "type": "ol",
            "q_mm2": 490,
            "alpha": alpha_al,
         "voltage_rating": "HV"},
        "679-AL1/86-ST1A 110.0":
        {"c_nf_per_km": 9.95,
            "r_ohm_per_km": 0.042,
            "x_ohm_per_km": 0.36,
            "max_i_ka": 1.150,
            "type": "ol",
            "q_mm2": 679,
            "alpha": alpha_al,
         "voltage_rating": "HV"},

        # Transmission System
        # The following values of c and x depend on the geometries of the  overhead line
        # Here it is assumed that for x the 220kV line uses twin conductors and the 380kV line uses
        # quad bundle conductor. The c values are estimated.
        "490-AL1/64-ST1A 220.0":
        {"c_nf_per_km": 10,
         "r_ohm_per_km": 0.059,
         "x_ohm_per_km": 0.285,
         "max_i_ka": 0.96,
         "type": "ol",
         "q_mm2": 490,
         "alpha": alpha_al,
         "voltage_rating": "HV"},
        "679-AL1/86-ST1A 220.0":
        {"c_nf_per_km": 11.7,
         "r_ohm_per_km": 0.042,
         "x_ohm_per_km": 0.275,
         "max_i_ka": 1.150,
         "type": "ol",
         "q_mm2": 679,
         "alpha": alpha_al,
         "voltage_rating": "HV"},
        "490-AL1/64-ST1A 380.0":
        {"c_nf_per_km": 11,
         "r_ohm_per_km": 0.059,
         "x_ohm_per_km": 0.253,
         "max_i_ka": 0.96,
         "type": "ol",
         "q_mm2": 490,
         "alpha": alpha_al,
         "voltage_rating": "HV"},
        "679-AL1/86-ST1A 380.0":
        {"c_nf_per_km": 14.6,
         "r_ohm_per_km": 0.042,
         "x_ohm_per_km": 0.25,
         "max_i_ka": 1.150,
         "type": "ol",
         "q_mm2": 679,
         "alpha": alpha_al,
         "voltage_rating": "HV"}
    }
    return linetypes


def basic_line_dc_std_types():

    alpha_al = 4.03e-3
    alpha_cu = 3.93e-3

    linedctypes = {
        # Cables, all from ABB-HVDC-Light--High-Voltage-Direct-Current--Cable-Catalogue.pdf
        # at the catalogue there are Land and submarine hvdc cables

        # High Voltage Submarine DC Bipolar Cupper Cables

         "95-CU":
        {"r_ohm_per_km": 0.193,
            "max_i_ka": 0.404,
            "type": "cs",
            "q_mm2": 95,
            "alpha": alpha_cu},
        "400-CU":
        {"r_ohm_per_km": 0.0470,
            "max_i_ka": 0.922,
            "type": "cs",
            "q_mm2": 400,
            "alpha": alpha_cu},
        "1200-CU":
        {"r_ohm_per_km": 0.0151 ,
            "max_i_ka": 1.791,
            "type": "cs",
            "q_mm2": 1200,
            "alpha": alpha_cu},
        "2400-CU":
        {"r_ohm_per_km": 0.0073,
            "max_i_ka": 2.678,
            "type": "cs",
            "q_mm2": 2400,
            "alpha": alpha_cu},

        # High Voltage Land DC Bipolar Aluminium Cables

        "95-AL":
            {"r_ohm_per_km": 0.32,
             "max_i_ka": 0.310,
             "type": "cs",
             "q_mm2": 95,
             "alpha": alpha_al},
        "400-AL":
            {"r_ohm_per_km": 0.0778,
             "max_i_ka": 0.705,
             "type": "cs",
             "q_mm2": 400,
             "alpha": alpha_al},
        "1200-AL":
            {"r_ohm_per_km": 0.0247,
             "max_i_ka": 1.371,
             "type": "cs",
             "q_mm2": 1200,
             "alpha": alpha_al},
        "2400-AL":
            {"r_ohm_per_km": 0.0121,
             "max_i_ka": 2.066,
             "type": "cs",
             "q_mm2": 2400,
             "alpha": alpha_al},
    }
    return linedctypes


def basic_trafo_std_types():
    trafotypes = {
        # derived from Oswald - Transformatoren - Vorlesungsskript Elektrische Energieversorgung I
        # another recommendable references for distribution transformers is Werth:
        # Netzberechnung mit Erzeugungsprofilen
        "160 MVA 380/110 kV":
        {"i0_percent": 0.06,
            "pfe_kw": 60,
            "vkr_percent": 0.25,
            "sn_mva": 160,
            "vn_lv_kv": 110.0,
            "vn_hv_kv": 380.0,
            "vk_percent": 12.2,
            "shift_degree": 0,
            "vector_group": "Yy0",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -9,
            "tap_max": 9,
            "tap_step_degree": 0,
            "tap_step_percent": 1.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
        "100 MVA 220/110 kV":
        {"i0_percent": 0.06,
            "pfe_kw": 55,
            "vkr_percent": 0.26,
            "sn_mva": 100,
            "vn_lv_kv": 110.0,
            "vn_hv_kv": 220.0,
            "vk_percent": 12.0,
            "shift_degree": 0,
            "vector_group": "Yy0",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -9,
            "tap_max": 9,
            "tap_step_degree": 0,
            "tap_step_percent": 1.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},

        # compare to IFT Ingenieurbüro data and Schlabbach book
        "63 MVA 110/20 kV":
        {"i0_percent": 0.04,
            "pfe_kw": 22,
            "vkr_percent": 0.32,
            "sn_mva": 63,
            "vn_lv_kv": 20.0,
            "vn_hv_kv": 110.0,
            "vk_percent": 18,
            "shift_degree": 150,
            "vector_group": "YNd5",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -9,
            "tap_max": 9,
            "tap_step_degree": 0,
            "tap_step_percent": 1.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
        "40 MVA 110/20 kV":
        {"i0_percent": 0.05,
            "pfe_kw": 18,
            "vkr_percent": 0.34,
            "sn_mva": 40,
            "vn_lv_kv": 20.0,
            "vn_hv_kv": 110.0,
            "vk_percent": 16.2,
            "shift_degree": 150,
            "vector_group": "YNd5",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -9,
            "tap_max": 9,
            "tap_step_degree": 0,
            "tap_step_percent": 1.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
        "25 MVA 110/20 kV":
        {"i0_percent": 0.07,
            "pfe_kw": 14,
            "vkr_percent": 0.41,
            "sn_mva": 25,
            "vn_lv_kv": 20.0,
            "vn_hv_kv": 110.0,
            "vk_percent": 12,
            "shift_degree": 150,
            "vector_group": "YNd5",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -9,
            "tap_max": 9,
            "tap_step_degree": 0,
            "tap_step_percent": 1.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
        "63 MVA 110/10 kV":
        {"sn_mva": 63,
            "vn_hv_kv": 110,
            "vn_lv_kv": 10,
            "vk_percent": 18,
            "vkr_percent": 0.32,
            "pfe_kw": 22,
            "i0_percent": 0.04,
            "shift_degree": 150,
            "vector_group": "YNd5",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -9,
            "tap_max": 9,
            "tap_step_degree": 0,
            "tap_step_percent": 1.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
        "40 MVA 110/10 kV":
        {"sn_mva": 40,
            "vn_hv_kv": 110,
            "vn_lv_kv": 10,
            "vk_percent": 16.2,
            "vkr_percent": 0.34,
            "pfe_kw": 18,
            "i0_percent": 0.05,
            "shift_degree": 150,
            "vector_group": "YNd5",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -9,
            "tap_max": 9,
            "tap_step_degree": 0,
            "tap_step_percent": 1.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
        "25 MVA 110/10 kV":
        {"sn_mva": 25,
            "vn_hv_kv": 110,
            "vn_lv_kv": 10,
            "vk_percent": 12,
            "vkr_percent": 0.41,
            "pfe_kw": 14,
            "i0_percent": 0.07,
            "shift_degree": 150,
            "vector_group": "YNd5",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -9,
            "tap_max": 9,
            "tap_step_degree": 0,
            "tap_step_percent": 1.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
        # Trafo 20/0.4
        # 0.25 MVA 20/0.4 kV 0.45 Trafo Union
        "0.25 MVA 20/0.4 kV":
        {"sn_mva": 0.25,
            "vn_hv_kv": 20,
            "vn_lv_kv": 0.4,
            "vk_percent": 6,
            "vkr_percent": 1.44,
            "pfe_kw": 0.8,
            "i0_percent": 0.32,
            "shift_degree": 150,
            "vector_group": "Yzn5",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -2,
            "tap_max": 2,
            "tap_step_degree": 0,
            "tap_step_percent": 2.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
        # 0.4 MVA 20/0.4 kV Trafo Union
        "0.4 MVA 20/0.4 kV":
        {"sn_mva": 0.4, "vn_hv_kv": 20, "vn_lv_kv": 0.4,
            "vk_percent": 6,
            "vkr_percent": 1.425,
            "pfe_kw": 1.35,
            "i0_percent": 0.3375,
            "shift_degree": 150,
            "vector_group": "Dyn5",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -2,
            "tap_max": 2,
            "tap_step_degree": 0,
            "tap_step_percent": 2.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
        # 0.63 MVA 20/0.4 kV Trafo Union
        "0.63 MVA 20/0.4 kV":
        {"sn_mva": 0.63,
            "vn_hv_kv": 20,
            "vn_lv_kv": 0.4,
            "vk_percent": 6,
            "vkr_percent": 1.206,
            "pfe_kw": 1.65,
            "i0_percent": 0.2619,
            "shift_degree": 150,
            "vector_group": "Dyn5",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -2,
            "tap_max": 2,
            "tap_step_degree": 0,
            "tap_step_percent": 2.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
        # Trafo 10/0.4:
        # 0.25 MVA 10/0.4 kV 0.4 Trafo Union wnr
        "0.25 MVA 10/0.4 kV":
        {"sn_mva": 0.25,
            "vn_hv_kv": 10,
            "vn_lv_kv": 0.4,
            "vk_percent": 4,
            "vkr_percent": 1.2,
            "pfe_kw": 0.6,
            "i0_percent": 0.24,
            "shift_degree": 150,
            "vector_group": "Dyn5",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -2,
            "tap_max": 2,
            "tap_step_degree": 0,
            "tap_step_percent": 2.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
        # 0.4 MVA 10/0.4 kV Trafo Union wnr
        "0.4 MVA 10/0.4 kV":
        {"sn_mva": 0.4,
            "vn_hv_kv": 10,
            "vn_lv_kv": 0.4,
            "vk_percent": 4,
            "vkr_percent": 1.325,
            "pfe_kw": 0.95,
            "i0_percent": 0.2375,
            "shift_degree": 150,
            "vector_group": "Dyn5",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -2,
            "tap_max": 2,
            "tap_step_degree": 0,
            "tap_step_percent": 2.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
        # 0.63 MVA 10/0.4 kV Trafo Union wnr
        "0.63 MVA 10/0.4 kV":
        {"sn_mva": 0.63,
            "vn_hv_kv": 10,
            "vn_lv_kv": 0.4,
            "vk_percent": 4,
            "vkr_percent": 1.0794,
            "pfe_kw": 1.18,
            "i0_percent": 0.1873,
            "shift_degree": 150,
            "vector_group": "Dyn5",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -2,
            "tap_max": 2,
            "tap_step_degree": 0,
            "tap_step_percent": 2.5,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
    }
    return trafotypes


def basic_trafo3w_std_types():
    trafo3wtypes = {
        # generic trafo3w
        "63/25/38 MVA 110/20/10 kV":
        {"sn_hv_mva": 63,
            "sn_mv_mva": 25,
            "sn_lv_mva": 38,
            "vn_hv_kv": 110,
            "vn_mv_kv": 20,
            "vn_lv_kv": 10,
            "vk_hv_percent": 10.4,
            "vk_mv_percent": 10.4,
            "vk_lv_percent": 10.4,
            "vkr_hv_percent": 0.28,
            "vkr_mv_percent": 0.32,
            "vkr_lv_percent": 0.35,
            "pfe_kw": 35,
            "i0_percent": 0.89,
            "shift_mv_degree": 0,
            "shift_lv_degree": 0,
            "vector_group": "YN0yn0yn0",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -10,
            "tap_max": 10,
            "tap_step_percent": 1.2,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False},
        "63/25/38 MVA 110/10/10 kV":
        {"sn_hv_mva": 63,
            "sn_mv_mva": 25,
            "sn_lv_mva": 38,
            "vn_hv_kv": 110,
            "vn_mv_kv": 10,
            "vn_lv_kv": 10,
            "vk_hv_percent": 10.4,
            "vk_mv_percent": 10.4,
            "vk_lv_percent": 10.4,
            "vkr_hv_percent": 0.28,
            "vkr_mv_percent": 0.32,
            "vkr_lv_percent": 0.35,
            "pfe_kw": 35,
            "i0_percent": 0.89,
            "shift_mv_degree": 0,
            "shift_lv_degree": 0,
            "vector_group": "YN0yn0yn0",
            "tap_side": "hv",
            "tap_neutral": 0,
            "tap_min": -10,
            "tap_max": 10,
            "tap_step_percent": 1.2,
            "tap_changer_type": "Ratio",
            "trafo_characteristic_table": False}
    }
    return trafo3wtypes


def basic_fuse_std_types():
    fusetypes = {
        'HV 100A': {'fuse_type': 'HV 100A',
                'i_rated_a': 100.0,
                't_avg': 0,
                't_min': [10.0, 3.64, 0.854, 0.281, 0.1, 0.0531, 0.022, 0.01],
                't_total': [10.0, 4.267, 1.21, 0.403, 0.1, 0.058, 0.022, 0.01],
                'x_avg': 0,
                'x_min': [300.0, 350.0, 450.0, 550.0, 700.0, 850.0, 1200.0, 1752.0],
                'x_total': [600.0, 700.0, 900.0, 1150.0, 1665.0, 2000.0, 3000.0, 4313.0]},
        'HV 10A': {'fuse_type': 'HV 10A',
                'i_rated_a': 10.0,
                't_avg': 0,
                't_min': [10.0, 1675.0, 0.344, 0.156, 0.1, 0.0417, 0.0171, 0.01],
                't_total': [10.0, 1.3, 0.3, 0.155, 0.1, 0.0555, 0.023, 0.01],
                'x_avg': 0,
                'x_min': [30.0, 32.0, 35.0, 37.0, 39.0, 50.0, 70.0, 88.0],
                'x_total': [60.0, 70.0, 80.0, 87.0, 94.0, 110.0, 150.0, 216.0]},
        'HV 125A': {'fuse_type': 'HV 125A',
                 'i_rated_a': 125.0,
                 't_avg': 0,
                 't_min': [10.0, 1.82, 0.344, 0.1, 0.0467, 0.0269, 0.01],
                 't_total': [10.0, 2.478, 0.426, 0.1, 0.0427, 0.0211, 0.01],
                 'x_avg': 0,
                 'x_min': [375.0, 500.0, 700.0, 925.0, 1200.0, 1500.0, 2341.0],
                 'x_total': [750.0, 1000.0, 1500.0, 2200.0, 3000.0, 4000.0, 5765.0]},
        'HV 160A': {'fuse_type': 'HV 160A',
                 'i_rated_a': 160.0,
                 't_avg': 0,
                 't_min': [10.0, 4.15, 1.03, 0.198, 0.1, 0.051, 0.0172, 0.01],
                 't_total': [10.0, 2.3, 0.734, 0.274, 0.1, 0.046, 0.0177, 0.01],
                 'x_avg': 0,
                 'x_min': [480.0, 550.0, 700.0, 1000.0, 1260.0, 1600.0, 2500.0, 3227.0],
                 'x_total': [960.0, 1300.0, 1700.0, 2200.0, 2996.0, 4000.0, 6000.0, 7946.0]},
        'HV 16A': {'fuse_type': 'HV 16A',
                'i_rated_a': 16.0,
                't_avg': 0,
                't_min': [10.0, 0.352, 0.164, 0.1, 0.0649, 0.0342, 0.01],
                't_total': [10.0, 2.34, 0.722, 0.181, 0.1, 0.055, 0.0296, 0.01],
                'x_avg': 0,
                'x_min': [48.0, 60.0, 65.0, 71.0, 80.0, 100.0, 162.0],
                'x_total': [96.0, 110.0, 125.0, 150.0, 168.0, 200.0, 250.0, 398.0]},
        'HV 200A': {'fuse_type': 'HV 200A',
                 'i_rated_a': 200.0,
                 't_avg': 0,
                 't_min': [10.0, 4.267, 1.21, 0.403, 0.1, 0.058, 0.022, 0.01],
                 't_total': [10.0, 3.73, 1.654, 0.328, 0.1, 0.0531, 0.019, 0.01],
                 'x_avg': 0,
                 'x_min': [600.0, 700.0, 900.0, 1150.0, 1665.0, 2000.0, 3000.0, 4313.0],
                 'x_total': [1200.0, 1500.0, 1800.0, 2700.0, 3960.0, 5000.0, 7500.0, 10620.0]},
        'HV 20A': {'fuse_type': 'HV 20A',
                'i_rated_a': 20.0,
                't_avg': 0,
                't_min': [10.0, 1.3, 0.3, 0.155, 0.1, 0.0555, 0.023, 0.01],
                't_total': [10.0, 1.3, 0.161, 0.1, 0.0611, 0.0399, 0.0141, 0.01],
                'x_avg': 0,
                'x_min': [60.0, 70.0, 80.0, 87.0, 94.0, 110.0, 150.0, 216.0],
                'x_total': [120.0, 150.0, 200.0, 223.0, 260.0, 300.0, 450.0, 532.0]},
        'HV 25A': {'fuse_type': 'HV 25A',
                'i_rated_a': 25.0,
                't_avg': 0,
                't_min': [10.0, 2.512, 0.833, 0.299, 0.1, 0.0372, 0.0223, 0.01],
                't_total': [10.0, 3.125, 0.597, 0.198, 0.1, 0.0378, 0.022, 0.01],
                'x_avg': 0,
                'x_min': [75.0, 82.0, 90.0, 100.0, 124.0, 170.0, 200.0, 289.0],
                'x_total': [150.0, 170.0, 210.0, 250.0, 294.0, 400.0, 500.0, 711.0]},
        'HV 31.5A': {'fuse_type': 'HV 31.5A',
                  'i_rated_a': 31.5,
                  't_avg': 0,
                  't_min': [10.0, 2.34, 0.722, 0.181, 0.1, 0.055, 0.0296, 0.01],
                  't_total': [10.0, 2.84, 0.368, 0.164, 0.1, 0.0621, 0.0378, 0.0195, 0.01],
                  'x_avg': 0,
                  'x_min': [95.0, 110.0, 125.0, 150.0, 165.0, 200.0, 250.0, 390.0],
                  'x_total': [189.0, 220.0, 300.0, 350.0, 393.0, 450.0, 530.0, 700.0, 960.0]},
        'HV 40A': {'fuse_type': 'HV 40A',
                'i_rated_a': 40.0,
                't_avg': 0,
                't_min': [10.0, 1.3, 0.161, 0.1, 0.0611, 0.0399, 0.0141, 0.01],
                't_total': [10.0, 2.05, 0.369, 0.198, 0.1, 0.051, 0.0298, 0.01],
                'x_avg': 0,
                'x_min': [120.0, 150.0, 200.0, 223.0, 260.0, 300.0, 450.0, 532.0],
                'x_total': [240.0, 300.0, 400.0, 450.0, 530.0, 650.0, 800.0, 1311.0]},
        'HV 50A': {'fuse_type': 'HV 50A',
                'i_rated_a': 50.0,
                't_avg': 0,
                't_min': [10.0, 3.215, 0.597, 0.198, 0.1, 0.0378, 0.022, 0.01],
                't_total': [10.0, 3.64, 0.854, 0.281, 0.1, 0.0531, 0.022, 0.01],
                'x_avg': 0,
                'x_min': [150.0, 170.0, 210.0, 250.0, 294.0, 400.0, 500.0, 711.0],
                'x_total': [300.0, 350.0, 450.0, 550.0, 700.0, 850.0, 1200.0, 1752.0]},
        'HV 6.3A': {'fuse_type': 'HV 6.3A',
                 'i_rated_a': 6.3,
                 't_avg': 0,
                 't_min': [10.0, 1.39, 0.344, 0.168, 0.1, 0.056, 0.0263, 0.01],
                 't_total': [10.0, 1.711, 0.516, 0.198, 0.1, 0.0634, 0.0303, 0.01],
                 'x_avg': 0,
                 'x_min': [19.0, 19.5, 20.4, 20.8, 22.0, 25.0, 32.0, 48.0],
                 'x_total': [38.0, 40.0, 43.0, 48.0, 53.0, 60.0, 75.0, 118.0]},
        'HV 63A': {'fuse_type': 'HV 63A',
                'i_rated_a': 63.0,
                't_avg': 0,
                't_min': [10.0, 2.84, 0.368, 0.164, 0.1, 0.0621, 0.0378, 0.0195, 0.01],
                't_total': [10.0, 1.82, 0.344, 0.1, 0.0467, 0.0269, 0.01],
                'x_avg': 0,
                'x_min': [189.0, 220.0, 300.0, 350.0, 393.0, 450.0, 530.0, 700.0, 961.0],
                'x_total': [378.0, 500.0, 700.0, 934.0, 1200.0, 1500.0, 2366.0]},
        'HV 80A': {'fuse_type': 'HV 80A',
                'i_rated_a': 80.0,
                't_avg': 0,
                't_min': [10.0, 2.05, 0.369, 0.198, 0.1, 0.051, 0.0298, 0.01],
                't_total': [10.0, 4.15, 1.03, 0.198, 0.1, 0.051, 0.0172, 0.01],
                'x_avg': 0,
                'x_min': [240.0, 300.0, 400.0, 450.0, 530.0, 650.0, 800.0, 1311.0],
                'x_total': [480.0, 550.0, 700.0, 1000.0, 1260.0, 1600.0, 2500.0, 3227.0]},
        'Siemens NH-1-100': {'fuse_type': 'Siemens NH-1-100',
                'i_rated_a': 100.0,
                't_avg': [5400.0, 2000.0, 400.0, 20.0, 1.0, 0.2, 0.012, 0.004],
                't_min': 0,
                't_total': 0,
                'x_avg': [150.0, 190.0, 250.0, 430.0, 900.0, 1250.0, 2700.0, 3600.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-1-125': {'fuse_type': 'Siemens NH-1-125',
                'i_rated_a': 125.0,
                't_avg': [4800.0, 120.0, 7.0, 0.1, 0.004],
                't_min': 0,
                't_total': 0,
                'x_avg': [180.0, 400.0, 740.0, 2000.0, 4250.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-1-16': {'fuse_type': 'Siemens NH-1-16',
                 'i_rated_a': 16.0,
                 't_avg': [4000.0, 400.0, 2.0, 0.1, 0.04, 0.01],
                 't_min': 0,
                 't_total': 0,
                 'x_avg': [26.0, 35.0, 75.0, 150.0, 200.0, 300.0],
                 'x_min': 0,
                 'x_total': 0},
        'Siemens NH-1-160': {'fuse_type': 'Siemens NH-1-160',
                'i_rated_a': 160.0,
                't_avg': [4800.0, 120.0, 7.0, 0.1, 0.004],
                't_min': 0,
                't_total': 0,
                'x_avg': [210.0, 500.0, 900.0, 2300.0, 5000.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-1-25': {'fuse_type': 'Siemens NH-1-25',
                'i_rated_a': 25.0,
                't_avg': [4000.0, 1000.0, 10.0, 0.2, 0.02, 0.01],
                't_min': 0,
                't_total': 0,
                'x_avg': [40.0, 50.0, 100.0, 210.0, 400.0, 500.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-1-50': {'fuse_type': 'Siemens NH-1-50',
                'i_rated_a': 50.0,
                't_avg': [4000.0, 40.0, 4.0, 1.0, 0.02, 0.01],
                't_min': 0,
                't_total': 0,
                'x_avg': [86.0, 200.0, 300.0, 400.0, 1000.0, 1280.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-1-63': {'fuse_type': 'Siemens NH-1-63',
                 'i_rated_a': 63.0,
                 't_avg': [4000.0, 100.0, 10.0, 2.0, 0.04, 0.01],
                 't_min': 0,
                 't_total': 0,
                 'x_avg': [100.0, 200.0, 300.0, 400.0, 1000.0, 1500.0],
                 'x_min': 0,
                 'x_total': 0},
        'Siemens NH-1-80': {'fuse_type': 'Siemens NH-1-80',
                'i_rated_a': 80.0,
                't_avg': [4800.0, 120.0, 7.0, 0.1, 0.01],
                't_min': 0,
                't_total': 0,
                'x_avg': [150.58, 250.0, 450.0, 1150.0, 2470.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-2-1000': {'fuse_type': 'Siemens NH-2-1000',
                'i_rated_a': 1000.0,
                't_avg': [4800.0, 120.0, 7.0, 0.1, 0.004],
                't_min': 0,
                't_total': 0,
                'x_avg': [1900.0, 3500.0, 8400.0, 24000.0, 52000.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-2-200': {'fuse_type': 'Siemens NH-2-200',
                'i_rated_a': 200.0,
                't_avg': [4800.0, 120.0, 7.0, 0.1, 0.004],
                't_min': 0,
                't_total': 0,
                'x_avg': [280.0, 650.0, 1200.0, 3000.0, 7000.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-2-224': {'fuse_type': 'Siemens NH-2-224',
                'i_rated_a': 224.0,
                't_avg': [4800.0, 120.0, 7.0, 0.2, 0.04, 0.004],
                't_min': 0,
                't_total': 0,
                'x_avg': [400.0, 750.0, 1453.0, 3025.0, 4315.0, 7600.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-2-250': {'fuse_type': 'Siemens NH-2-250',
                'i_rated_a': 250.0,
                't_avg': [4800.0, 120.0, 7.0, 0.1, 0.004],
                't_min': 0,
                't_total': 0,
                'x_avg': [450.0, 800.0, 1650.0, 4000.0, 8500.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-2-315': {'fuse_type': 'Siemens NH-2-315',
                'i_rated_a': 315.0,
                't_avg': [4800.0, 120.0, 7.0, 0.1, 0.004],
                't_min': 0,
                't_total': 0,
                'x_avg': [550.0, 920.0, 1900.0, 5000.0, 11000.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-2-355': {'fuse_type': 'Siemens NH-2-355',
                'i_rated_a': 355.0,
                't_avg': [4800.0, 120.0, 6.0, 0.1, 0.004],
                't_min': 0,
                't_total': 0,
                'x_avg': [650.0, 1116.27, 2350.0, 5840.0, 12790.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-2-400': {'fuse_type': 'Siemens NH-2-400',
                'i_rated_a': 400.0,
                't_avg': [4800.0, 120.0, 7.0, 0.1, 0.004],
                't_min': 0,
                't_total': 0,
                'x_avg': [720.0, 1350.0, 2800.0, 6500.0, 15000.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-2-425': {'fuse_type': 'Siemens NH-2-425',
                'i_rated_a': 425.0,
                't_avg': [4800.0, 120.0, 7.0, 0.1, 0.004],
                't_min': 0,
                't_total': 0,
                'x_avg': [850.0, 1500.0, 3050.0, 7500.0, 16500.0],
                'x_min': 0,
                'x_total': 0},
        'Siemens NH-2-630': {'fuse_type': 'Siemens NH-2-630',
                'i_rated_a': 630.0,
                't_avg': [4800.0, 120.0, 7.0, 0.1, 0.004],
                't_min': 0,
                't_total': 0,
                'x_avg': [1200.0, 2000.0, 4800.0, 12000.0, 26000.0],
                'x_min': 0,
                'x_total': 0}
    }
    return fusetypes


def basic_std_types():
    return {
        "line"   : basic_line_std_types(),
        "line_dc"   : basic_line_dc_std_types(),
        "trafo"  : basic_trafo_std_types(),
        "trafo3w": basic_trafo3w_std_types(),
        "fuse"   : basic_fuse_std_types()
    }


def add_basic_std_types(net):
    """Adds basic standard types of the pandapower library to the net provided. These standard types
    are the same types that are available with output of `pandapower.create_empty_network()` and
    `pandapower.create_empty_network(add_stdtypes=True)` respectively.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net which should receive the basic standard types

    Returns
    -------
    tuple of dictionaries
        line,line_dc, trafo and trafo3w types as dictionaries which have been added to the net.
    """

    if "std_types" not in net:
        net.std_types = {"line": {}, "line_dc": {}, "trafo": {}, "trafo3w": {}, "fuse": {}}

    linetypes = basic_line_std_types()
    linedctypes = basic_line_dc_std_types()
    trafotypes = basic_trafo_std_types()
    trafo3wtypes = basic_trafo3w_std_types()
    fusetypes = basic_fuse_std_types()

    create_std_types(net, data=linetypes, element="line")
    create_std_types(net, data=linedctypes, element="line_dc")
    create_std_types(net, data=trafotypes, element="trafo")
    create_std_types(net, data=trafo3wtypes, element="trafo3w")
    create_std_types(net, data=fusetypes, element="fuse")
    return linetypes, linedctypes, trafotypes, trafo3wtypes, fusetypes



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from pandapower.pypower.idx_brch import branch_cols as start

# columns for TDPF

TDPF 						= start + 0   # bool flag (True or False) of whether the branch must be considered in the TDPF power flow
BR_R_REF_OHM_PER_KM 		= start + 1   # reference R is needed because BR_R in ppc can already be adjusted by temperature
BR_LENGTH_KM 				= start + 3   # to calculate R_ref in p.u.
RATE_I_KA 					= start + 4   # rated I limit in kA (max_i_ka)
T_START_C 					= start + 5   # starting temperature of the conductor in °C
T_REF_C 					= start + 6   # reference temperature for calculating updated R (optional)
T_AMBIENT_C 				= start + 7   # temperature of the surrounding air
ALPHA 						= start + 8   # thermal coefficient of resistance (4.03e-3 for Al)
WIND_SPEED_MPS 				= start + 9   # Wind speed in meters per second
WIND_ANGLE_DEGREE 			= start + 10  # Angle of attack, angle between wind direction and conductor
SOLAR_RADIATION_W_PER_SQ_M 	= start + 11  # Solar radiation in W/m²
GAMMA 						= start + 12  # absorption albedo, solar absorptivity
EPSILON 					= start + 13  # radiative albedo, emissivity
R_THETA 					= start + 14  # Thermal resistance R_Theta in K/MW (Optional)
OUTER_DIAMETER_M 			= start + 15  # outer diameter of the conductor
MC_JOULE_PER_M_K 			= start + 16  # thermal inertia parameter of the conductor

branch_cols_tdpf 			= 17



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.



IM_FROM = 0
IM_FROM_STD = 1
IM_TO = 2
IM_TO_STD = 3
P_FROM = 4
P_FROM_STD = 5
P_TO = 6
P_TO_STD = 7
Q_FROM = 8
Q_FROM_STD = 9
Q_TO = 10
Q_TO_STD = 11
IM_FROM_IDX = 12
P_FROM_IDX = 13
Q_FROM_IDX = 14
IM_TO_IDX = 15
P_TO_IDX = 16
Q_TO_IDX = 17
IA_FROM = 18
IA_FROM_STD = 19
IA_TO = 20
IA_TO_STD = 21
IA_FROM_IDX = 22
IA_TO_IDX = 23


branch_cols_se = 24



# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


"""Solves the power flow using a full Newton's method.
"""
import numpy as np
from numpy import float64, array, angle, sqrt, square, exp, linalg, conj, r_, inf, arange, zeros, \
    max, zeros_like, column_stack, flatnonzero, nan_to_num
from pandapower.pypower.bustypes import bustypes_dc
from pandapower.pypower.idx_brch_dc import DC_BR_R, DC_PF, DC_IF, DC_PT, DC_IT, DC_BR_STATUS, DC_F_BUS, DC_T_BUS
from scipy.sparse import csr_matrix, eye, vstack
from scipy.sparse.linalg import spsolve

from pandapower.auxiliary import _sum_by_group
from pandapower.pf.iwamoto_multiplier import _iwamoto_step
from pandapower.pf.makeYbus_facts import makeYbus_svc, makeYft_tcsc, calc_y_svc_pu, \
    makeYbus_ssc_vsc, make_Ybus_facts, make_Yft_facts
from pandapower.pypower.idx_bus_dc import DC_PD, DC_VM, DC_BUS_TYPE, DC_NONE, DC_BUS_I, DC_REF, DC_P
from pandapower.pypower.idx_vsc import VSC_CONTROLLABLE, VSC_MODE_AC, VSC_VALUE_AC, VSC_MODE_DC, VSC_VALUE_DC, VSC_R, \
    VSC_X, VSC_Q, VSC_P, VSC_BUS_DC, VSC_P_DC, VSC_MODE_AC_SL, VSC_MODE_AC_V, VSC_MODE_AC_Q, VSC_MODE_DC_P, \
    VSC_MODE_DC_V, VSC_INTERNAL_BUS_DC, VSC_R_DC, VSC_PL_DC, VSC_STATUS, VSC_BUS, VSC_INTERNAL_BUS
from pandapower.pypower.makeSbus import makeSbus
from pandapower.pf.create_jacobian import create_jacobian_matrix, get_fastest_jacobian_function
from pandapower.pypower.idx_gen import PG
from pandapower.pypower.idx_bus import PD, SL_FAC, BASE_KV
from pandapower.pypower.idx_brch import BR_R, BR_X, F_BUS, T_BUS
from pandapower.pypower.idx_brch_tdpf import BR_R_REF_OHM_PER_KM, BR_LENGTH_KM, RATE_I_KA, \
    T_START_C, R_THETA, WIND_SPEED_MPS, ALPHA, TDPF, OUTER_DIAMETER_M, MC_JOULE_PER_M_K, \
    WIND_ANGLE_DEGREE, SOLAR_RADIATION_W_PER_SQ_M, GAMMA, EPSILON, T_AMBIENT_C, T_REF_C
from pandapower.pypower.idx_tcsc import TCSC_F_BUS, TCSC_T_BUS, TCSC_X_L, TCSC_X_CVAR, TCSC_SET_P, \
    TCSC_THYRISTOR_FIRING_ANGLE, TCSC_STATUS, TCSC_CONTROLLABLE, TCSC_MIN_FIRING_ANGLE, \
    TCSC_MAX_FIRING_ANGLE, TCSC_PF, TCSC_QF, TCSC_PT, TCSC_QT, TCSC_IF, TCSC_IT, TCSC_X_PU
from pandapower.pypower.idx_svc import SVC_BUS, SVC_STATUS, SVC_CONTROLLABLE, SVC_X_L, SVC_X_CVAR, \
    SVC_X_PU, SVC_SET_VM_PU, SVC_THYRISTOR_FIRING_ANGLE, SVC_MAX_FIRING_ANGLE, \
    SVC_MIN_FIRING_ANGLE, SVC_Q
from pandapower.pypower.idx_ssc import SSC_BUS, SSC_R, SSC_X, SSC_SET_VM_PU, SSC_STATUS, \
    SSC_CONTROLLABLE, SSC_Q, SSC_X_CONTROL_VM, SSC_X_CONTROL_VA, SSC_INTERNAL_BUS

from pandapower.pf.create_jacobian_tdpf import calc_g_b, calc_a0_a1_a2_tau, calc_r_theta, \
    calc_T_frank, calc_i_square_p_loss, create_J_tdpf

from pandapower.pf.create_jacobian_facts import create_J_modification_svc, \
    create_J_modification_tcsc, create_J_modification_ssc_vsc, create_J_modification_hvdc


def newtonpf(Ybus, Sbus, V0, ref, pv, pq, ppci, options, makeYbus=None):
    """Solves the power flow using a full Newton's method.
    Solves for bus voltages given the full system admittance matrix (for
    all buses), the complex bus power injection vector (for all buses),
    the initial vector of complex bus voltages, and column vectors with
    the lists of bus indices for the swing bus, PV buses, and PQ buses,
    respectively. The bus voltage vector contains the set point for
    generator (including ref bus) buses, and the reference angle of the
    swing bus, as well as an initial guess for remaining magnitudes and
    angles.
    @see: L{runpf}
    @author: Ray Zimmerman (PSERC Cornell)
    @author: Richard Lincoln
    Modified by University of Kassel (Florian Schaefer) to use numba
    """

    # options
    tol = options['tolerance_mva']
    max_it = options["max_iteration"]
    numba = options["numba"]
    iwamoto = options["algorithm"] == "iwamoto_nr"
    voltage_depend_loads = options["voltage_depend_loads"]
    dist_slack = options["distributed_slack"]
    v_debug = options["v_debug"]
    use_umfpack = options["use_umfpack"]
    permc_spec = options["permc_spec"]

    baseMVA = ppci['baseMVA']
    bus = ppci['bus']
    bus_dc = ppci['bus_dc']
    gen = ppci['gen']
    branch = ppci['branch']
    branch_dc = ppci['branch_dc']
    tcsc = ppci['tcsc']
    svc = ppci['svc']
    ssc = ppci['ssc']
    vsc = ppci['vsc']
    slack_weights = bus[:, SL_FAC].astype(float64)  ## contribution factors for distributed slack
    tdpf = options.get('tdpf', False)

    # FACTS
    ## svc
    svc_idx = flatnonzero(nan_to_num(svc[:, SVC_STATUS]))
    svc_buses = svc[svc_idx, SVC_BUS].astype(np.int64)
    svc_set_vm_pu = svc[svc_idx, SVC_SET_VM_PU]
    x_control_svc = svc[svc_idx, SVC_THYRISTOR_FIRING_ANGLE]
    svc_controllable = svc[svc_idx, SVC_CONTROLLABLE].astype(bool)
    svc_x_l_pu = svc[svc_idx, SVC_X_L]
    svc_x_cvar_pu = svc[svc_idx, SVC_X_CVAR]
    svc_min_x = svc[svc_idx[svc_controllable], SVC_MIN_FIRING_ANGLE].real
    svc_max_x = svc[svc_idx[svc_controllable], SVC_MAX_FIRING_ANGLE].real
    num_svc_controllable = len(x_control_svc[svc_controllable])
    num_svc = len(x_control_svc)
    any_svc = num_svc > 0
    any_svc_controllable = num_svc_controllable > 0

    ## tcsc
    tcsc_branches = flatnonzero(nan_to_num(tcsc[:, TCSC_STATUS]))
    tcsc_fb = tcsc[tcsc_branches, [TCSC_F_BUS]].real.astype(np.int64)
    tcsc_tb = tcsc[tcsc_branches, [TCSC_T_BUS]].real.astype(np.int64)
    tcsc_controllable = tcsc[tcsc_branches, TCSC_CONTROLLABLE].real.astype(bool)
    tcsc_set_p_pu = tcsc[tcsc_branches[tcsc_controllable], TCSC_SET_P].real
    tcsc_min_x = tcsc[tcsc_branches[tcsc_controllable], TCSC_MIN_FIRING_ANGLE].real
    tcsc_max_x = tcsc[tcsc_branches[tcsc_controllable], TCSC_MAX_FIRING_ANGLE].real
    x_control_tcsc = tcsc[tcsc_branches, TCSC_THYRISTOR_FIRING_ANGLE].real
    tcsc_x_l_pu = tcsc[tcsc_branches, TCSC_X_L].real #### should not this be imag ?
    tcsc_x_cvar_pu = tcsc[tcsc_branches, TCSC_X_CVAR].real
    num_tcsc = len(x_control_tcsc)
    num_tcsc_controllable = len(x_control_tcsc[tcsc_controllable])
    any_tcsc = num_tcsc > 0
    any_tcsc_controllable = num_tcsc_controllable > 0

    ## ssc
    ssc_branches = flatnonzero(nan_to_num(ssc[:, SSC_STATUS]))
    ssc_fb = ssc[ssc_branches, [SSC_BUS]].real.astype(np.int64)
    ssc_tb = ssc[ssc_branches, [SSC_INTERNAL_BUS]].real.astype(np.int64)
    size_y = Ybus.shape[0]
    num_ssc = len(ssc_fb)
    ssc_controllable = ssc[ssc_branches, SSC_CONTROLLABLE].real.astype(bool)
    x_control_ssc = ssc[ssc_branches, SSC_X_CONTROL_VM].real * np.exp(1j * ssc[ssc_branches, SSC_X_CONTROL_VA].real)
    num_ssc_controllable = len(x_control_ssc[ssc_controllable])
    ssc_set_vm_pu = ssc[ssc_branches[ssc_controllable], SSC_SET_VM_PU]
    ssc_mode_ac = np.zeros_like(ssc_set_vm_pu)
    ssc_y_pu = 1/(ssc[ssc_branches, SSC_R].real + 1j * ssc[ssc_branches, SSC_X].real)
    any_ssc = num_ssc > 0
    any_ssc_controllable = num_ssc_controllable > 0

    ## vsc
    bus_dc[:, DC_BUS_TYPE] = DC_P # todo vsc
    vsc_branches = flatnonzero(nan_to_num(vsc[:, VSC_STATUS]))
    vsc_fb = vsc[vsc_branches, [VSC_BUS]].real.astype(np.int64)
    vsc_tb = vsc[vsc_branches, [VSC_INTERNAL_BUS]].real.astype(np.int64)
    vsc_dc_fb = vsc[vsc_branches, [VSC_BUS_DC]].real.astype(np.int64)
    vsc_dc_tb = vsc[vsc_branches, [VSC_INTERNAL_BUS_DC]].real.astype(np.int64)
    num_vsc = len(vsc_fb)
    vsc_controllable = vsc[vsc_branches, VSC_CONTROLLABLE].real.astype(bool)
    num_vsc_controllable = sum(vsc_controllable)
    vsc_mode_ac = vsc[vsc_branches[vsc_controllable], VSC_MODE_AC]
    vsc_value_ac = vsc[vsc_branches[vsc_controllable], VSC_VALUE_AC]
    vsc_mode_dc = vsc[vsc_branches[vsc_controllable], VSC_MODE_DC]
    vsc_value_dc = vsc[vsc_branches[vsc_controllable], VSC_VALUE_DC]
    vsc_y_pu = 1 / (vsc[vsc_branches, VSC_R] + 1j * vsc[vsc_branches, VSC_X].real)
    # make sure this is not 0 in the inputs!
    with np.errstate(all="raise"):
        vsc_g_pu = 1 / vsc[vsc_branches, VSC_R_DC]
    vsc_gl_pu = vsc[vsc_branches, VSC_PL_DC] # / np.square(bus_dc[vsc[vsc_branches, VSC_BUS_DC].astype(np.int64), DC_BASE_KV])
    any_vsc = num_vsc > 0
    any_vsc_controllable = num_vsc_controllable > 0
    vsc_dc_mode_p = (vsc_mode_dc == VSC_MODE_DC_P) & (vsc_mode_ac != VSC_MODE_AC_SL)
    vsc_dc_mode_v = (vsc_mode_dc == VSC_MODE_DC_V) & (vsc_mode_ac != VSC_MODE_AC_SL)

    hvdc_fb = branch_dc[:, DC_F_BUS].astype(np.int64)
    hvdc_tb = branch_dc[:, DC_T_BUS].astype(np.int64)

    baseR = 110 ** 2 / baseMVA
    hvdc_branches = np.flatnonzero(branch_dc[:, DC_BR_STATUS] != 0)
    num_branch_dc = len(hvdc_branches) + num_vsc  # internal vsc resistance is also a dc branch
    with np.errstate(all="raise"):
        hvdc_y_pu = 1 / branch_dc[hvdc_branches, DC_BR_R]
    relevant_bus_dc = flatnonzero(bus_dc[:, DC_BUS_TYPE] != DC_NONE)
    num_bus_dc = len(relevant_bus_dc)
    any_branch_dc = num_branch_dc > 0
    P_dc = -bus_dc[relevant_bus_dc, DC_PD]  # load is negative here
    p_set_point_index = vsc_controllable & (vsc_mode_dc == VSC_MODE_DC_P) & (vsc_mode_ac != VSC_MODE_AC_SL)
    # P_dc[vsc[p_set_point_index, VSC_BUS_DC].astype(np.int64)] = -vsc_value_dc[vsc_mode_dc == 1]  # todo sum by group
    # todo vsc
    vsc_group_buses_p, P_dc_sum, vsc_group_buses_p_number = _sum_by_group(vsc_dc_tb[p_set_point_index], -vsc_value_dc[p_set_point_index], np.ones(sum(p_set_point_index)))
    #P_dc[vsc_group_buses_p] = P_dc_sum
    vsc_slack_p_dc_bus, _, _ = _sum_by_group(vsc_dc_tb[vsc_mode_ac == VSC_MODE_AC_SL], vsc_dc_tb[vsc_mode_ac == VSC_MODE_AC_SL], vsc_dc_tb[vsc_mode_ac == VSC_MODE_AC_SL])
    P_dc_sum_sl = P_dc[vsc_slack_p_dc_bus].copy()  # later used in mismatch for vsc
    #vsc_group_buses_ref, _, vsc_group_buses_ref_number = _sum_by_group(vsc_dc_tb[p_set_point_index], -vsc_value_dc[vsc_mode_dc == 1], np.ones(sum(p_set_point_index)))

    # J for HVDC is expanded by the number of DC "P" buses (added below)
    num_facts_controllable = num_svc_controllable + num_tcsc_controllable # + 2 * num_ssc_controllable
    num_facts = num_svc + num_tcsc + num_ssc  # todo schould it be num_bus?


    #
    # tcsc_in_pq_f = np.isin(branch[tcsc_branches, F_BUS].real.astype(np.int64), pq)
    # tcsc_in_pq_t = np.isin(branch[tcsc_branches, T_BUS].real.astype(np.int64), pq)
    # tcsc_in_pvpq_f = np.isin(branch[tcsc_branches, F_BUS].real.astype(np.int64), pvpq)
    # tcsc_in_pvpq_t = np.isin(branch[tcsc_branches, T_BUS].real.astype(np.int64), pvpq)
    # # else:
    # #   tcsc_fb = tcsc_tb = tcsc_i = tcsc_j = None

    ############

    # initialize
    i = 0
    V = V0
    V_dc = np.ones(num_bus_dc, dtype=np.float64)  # initial voltage vector for the DC line
    v_set_point_index = vsc_controllable & (vsc_mode_dc == 0)
    if len(vsc_value_dc[v_set_point_index]) > 0:
        V_dc[:] = np.mean(vsc_value_dc[v_set_point_index])
        # V_dc[vsc[v_set_point_index, VSC_BUS_DC].astype(np.int64)] = vsc_value_dc[v_set_point_index]

    Ybus_svc = makeYbus_svc(Ybus, x_control_svc, svc_x_l_pu, svc_x_cvar_pu, svc_buses)
    y_tcsc_pu = -1j * calc_y_svc_pu(x_control_tcsc, tcsc_x_l_pu, tcsc_x_cvar_pu)
    Ybus_tcsc = make_Ybus_facts(tcsc_fb, tcsc_tb, y_tcsc_pu, Ybus.shape[0])
    # Ybus_tcsc = makeYbus_tcsc(Ybus, x_control_tcsc, tcsc_x_l_pu, tcsc_x_cvar_pu, tcsc_fb, tcsc_tb)
    # SSC
    Ybus_ssc_not_controllable, Ybus_ssc_controllable, Ybus_ssc = \
        makeYbus_ssc_vsc(Ybus, ssc_y_pu, ssc_fb, ssc_tb, ssc_controllable)
    # VSC
    Ybus_vsc_not_controllable, Ybus_vsc_controllable, Ybus_vsc = \
        makeYbus_ssc_vsc(Ybus, vsc_y_pu, vsc_fb, vsc_tb, vsc_controllable)
    # HVDC
    Ybus_vsc_dc = make_Ybus_facts(vsc_dc_fb, vsc_dc_tb, vsc_g_pu, num_bus_dc, ysf_pu=vsc_gl_pu, dtype=np.float64)
    Ybus_hvdc = make_Ybus_facts(hvdc_fb, hvdc_tb, hvdc_y_pu, num_bus_dc, dtype=np.float64) + Ybus_vsc_dc
    Yf_vsc_dc, Yt_vsc_dc = make_Yft_facts(vsc_dc_fb, vsc_dc_tb, vsc_g_pu, num_bus_dc, ysf_pu=vsc_gl_pu)
    Yf_vsc, Yt_vsc = make_Yft_facts(vsc_fb, vsc_tb, vsc_y_pu, Ybus_vsc.shape[0])

    # to avoid non-convergence due to zero-terms in the Jacobian:
    if any_tcsc_controllable and np.all(V[tcsc_fb] == V[tcsc_tb]):
        V[tcsc_tb] -= 0.01 + 0.001j
    # if any_ssc_controllable and np.all(V[ssc_fb[ssc_controllable]] == V[ssc_tb[ssc_controllable]]):
    #     V[ssc_tb[ssc_controllable]] -= 0.01 + 0.001j
    # if any_vsc_controllable and np.all(V[vsc_fb[vsc_controllable]] == V[vsc_tb[vsc_controllable]]):
    #     V[vsc_tb[vsc_controllable]] -= 0.01 + 0.001j

    Va = angle(V)
    Vm = abs(V)
    dVa, dVm = None, None
    if iwamoto:
        dVm, dVa = zeros_like(Vm), zeros_like(Va)

    if v_debug:
        Vm_it = Vm.copy()
        Va_it = Va.copy()
    else:
        Vm_it = None
        Va_it = None

    # set up indexing for updating V
    if dist_slack and len(ref) > 1:
        pv = r_[ref[1:], pv]
        ref = ref[[0]]

    pvpq = r_[pv, pq]
    # reference buses are always at the top, no matter where they are in the grid (very confusing...)
    # so in the refpvpq, the indices must be adjusted so that ref bus(es) starts with 0
    # todo: is it possible to simplify the indices/lookups and make the code clearer?
    # for columns: columns are in the normal order in Ybus; column numbers for J are reduced by 1 internally
    refpvpq = r_[ref, pvpq]
    # generate lookup pvpq -> index pvpq (used in createJ):
    #   shows for a given row from Ybus, which row in J it becomes
    #   e.g. the first row in J is a PV bus. If the first PV bus in Ybus is in the row 2, the index of the row in Jbus must be 0.
    #   pvpq_lookup will then have a 0 at the index 2
    pvpq_lookup = zeros(max((Ybus + Ybus_svc + Ybus_tcsc + Ybus_ssc + Ybus_vsc).indices) + 1, dtype=np.int64)
    if dist_slack:
        # slack bus is relevant for the function createJ_ds
        pvpq_lookup[refpvpq] = arange(len(refpvpq))
    else:
        pvpq_lookup[pvpq] = arange(len(pvpq))

    pq_lookup = zeros(max(refpvpq) + 1, dtype=np.int64)
    pq_lookup[pq] = arange(len(pq))

    # lookups for the DC system. For now, we assume the DC "from" bus is always slack
    dc_ref, dc_b2b, dc_p = bustypes_dc(bus_dc)
    num_dc_p = len(dc_p)
    dc_refb2bp = r_[dc_ref, dc_b2b, dc_p]
    if len(dc_refb2bp) > 0:
        dc_p_lookup = zeros(max(dc_refb2bp) + 1, dtype=np.int64)
        dc_p_lookup[dc_p] = arange(len(dc_p))
        dc_ref_lookup = zeros(max(dc_refb2bp) + 1, dtype=np.int64)
        dc_ref_lookup[dc_ref] = arange(len(dc_ref))
        dc_b2b_lookup = zeros(max(dc_refb2bp) + 1, dtype=np.int64)
        dc_b2b_lookup[dc_b2b] = arange(len(dc_b2b))
        num_facts_controllable += num_dc_p
    else:
        dc_p_lookup = np.array([], dtype=np.int64)
        dc_ref_lookup = np.array([], dtype=np.int64)
        dc_b2b_lookup = np.array([], dtype=np.int64)

    # get jacobian function
    createJ = get_fastest_jacobian_function(pvpq, pq, numba, dist_slack)

    nref = len(ref)
    npv = len(pv)
    npq = len(pq)
    # todo: tidy up the indices below
    j0 = 0
    j1 = nref if dist_slack else 0
    j2 = j1 + npv  # j1:j2 - V angle of pv buses
    j3 = j2
    j4 = j2 + npq  # j3:j4 - V angle of pq buses
    j5 = j4
    j6 = j4 + npq  # j5:j6 - V mag of pq buses
    j6a = j6 + num_svc_controllable  # svc
    j6b = j6a + num_tcsc_controllable  # tcsc
    j6c = j6b + 0 # num_ssc_controllable # ssc Vq angle
    j7 = j6c
    j6d = j6c + 0 # num_ssc_controllable # ssc Vq mag
    j6e = j6d + (num_dc_p if any_branch_dc else 0)  # V of the p-buses of the DC grid
    j8 = j6e

    # make initial guess for the slack
    slack = (gen[:, PG].sum() - bus[:, PD].sum()) / baseMVA
    # evaluate F(x0)
    any_facts_controllable = num_facts_controllable > 0
    F = _evaluate_Fx(Ybus + Ybus_svc + Ybus_tcsc + Ybus_ssc + Ybus_vsc, V, Sbus, ref, pv, pq, slack_weights, dist_slack, slack)
    if any_facts_controllable or any_ssc_controllable or any_vsc_controllable:
        mis_facts = _evaluate_Fx_facts(V, pq, svc_buses[svc_controllable], svc_set_vm_pu[svc_controllable],
                                       tcsc_controllable, tcsc_set_p_pu, tcsc_tb, Ybus_tcsc, ssc_fb[ssc_controllable],
                                       ssc_tb[ssc_controllable], Ybus_ssc, ssc_controllable, ssc_set_vm_pu, F,
                                       pq_lookup, vsc_controllable, vsc_fb, vsc_tb, Ybus_vsc, Ybus_vsc_dc, Yf_vsc_dc,
                                       Yt_vsc_dc, Yf_vsc, vsc_mode_ac, vsc_mode_dc, vsc_value_ac, vsc_value_dc,
                                       vsc_dc_fb, vsc_dc_tb, vsc_dc_mode_v, vsc_dc_mode_p, vsc_gl_pu, V_dc, Ybus_hvdc,
                                       num_branch_dc, P_dc, dc_p, dc_ref, dc_b2b, dc_p_lookup, dc_ref_lookup,
                                       dc_b2b_lookup, P_dc_sum_sl)
        F = r_[F, mis_facts]

    T_base = 100  # T in p.u. for better convergence
    T = 20 / T_base
    r_theta_pu = 0
    if tdpf:
        if len(pq) > 0:
            pq_lookup = zeros(max(refpvpq) + 1, dtype=np.int64)  # for TDPF
            pq_lookup[pq] = arange(len(pq))
        else:
            pq_lookup = array([])
        tdpf_update_r_theta = options.get('tdpf_update_r_theta', True)
        tdpf_delay_s = options.get('tdpf_delay_s')
        tdpf_lines = flatnonzero(nan_to_num(branch[:, TDPF]))
        # set up the necessary parameters for TDPF:
        T0 = branch[tdpf_lines, T_START_C].real / T_base
        t_ref_pu = branch[tdpf_lines, T_REF_C].real / T_base
        t_air_pu = branch[tdpf_lines, T_AMBIENT_C].real / T_base
        alpha_pu = branch[tdpf_lines, ALPHA].real * T_base

        i_max_a = branch[tdpf_lines, RATE_I_KA].real * 1e3
        v_base_kv = bus[branch[tdpf_lines, F_BUS].real.astype(np.int64), BASE_KV]
        z_base_ohm = square(v_base_kv) / baseMVA
        r_ref_pu = branch[tdpf_lines, BR_R_REF_OHM_PER_KM].real * branch[tdpf_lines, BR_LENGTH_KM].real / z_base_ohm
        i_base_a = baseMVA / (v_base_kv * sqrt(3)) * 1e3
        i_max_pu = i_max_a / i_base_a
        # p_rated_loss_pu = square(i_max_pu) * r_ref_pu * (1 + alpha_pu * (25/T_base+t_air_pu - t_ref_pu))
        # p_rated_loss_mw = square(branch[tdpf_lines, RATE_I_KA].real * sqrt(3)) * branch[tdpf_lines, BR_R_REF_OHM_PER_KM].real * branch[tdpf_lines, BR_LENGTH_KM].real * (1 + alpha_pu * (25/T_base+t_air_pu - t_ref_pu))
        # assert np.allclose(p_rated_loss_mw / baseMVA, p_rated_loss_pu)
        # defined in Frank et.al. as T_Rated_Rise / p_rated_loss. Given in net.line based on °C, kA, kV:
        r_theta_pu = branch[tdpf_lines, R_THETA].real * baseMVA / T_base
        x = branch[tdpf_lines, BR_X].real

        # calculate parameters for J:
        Ybus, Yf, Yt = makeYbus(baseMVA, bus, branch)
        # todo: add parameters to the create function
        a0, a1, a2, tau = calc_a0_a1_a2_tau(t_air_pu=t_air_pu, t_max_pu=80 / T_base, t_ref_pu=t_ref_pu,
                                            r_ref_ohm_per_m=1e-3 * branch[tdpf_lines, BR_R_REF_OHM_PER_KM].real,
                                            conductor_outer_diameter_m=branch[tdpf_lines, OUTER_DIAMETER_M].real,
                                            mc_joule_per_m_k=branch[tdpf_lines, MC_JOULE_PER_M_K].real,
                                            wind_speed_m_per_s=branch[tdpf_lines, WIND_SPEED_MPS].real,
                                            wind_angle_degree=branch[tdpf_lines, WIND_ANGLE_DEGREE].real,
                                            s_w_per_square_meter=branch[tdpf_lines, SOLAR_RADIATION_W_PER_SQ_M].real,
                                            alpha_pu=alpha_pu, solar_absorptivity=branch[tdpf_lines, GAMMA].real,
                                            emissivity=branch[tdpf_lines, EPSILON].real, T_base=T_base,
                                            i_base_a=i_base_a)
        g, b = calc_g_b(r_ref_pu, x)
        i_square_pu, p_loss_pu = calc_i_square_p_loss(branch, tdpf_lines, g, b, Vm, Va)
        if tdpf_update_r_theta:
            r_theta_pu = calc_r_theta(t_air_pu, a0, a1, a2, i_square_pu, p_loss_pu)
        # initial guess for T:
        # T = calc_T_frank(p_loss_pu, t_air_pu, r_theta_pu, tdpf_delay_s, T0, tau)
        T = T0.copy()  # better for e.g. timeseries calculation
        F_t = zeros(len(branch))
        # F_t[tdpf_lines] = T - T0
        F = r_[F, F_t]

    converged = _check_for_convergence(F, tol)

    Ybus = Ybus.tocsr()
    J = None


    # do Newton iterations
    while (not converged and i < max_it):
        # update iteration counter
        i = i + 1

        if tdpf:
            # update the R, g, b for the tdpf_lines, and the Y-matrices
            # here: limit the change of the R to reflect a realistic range of values for T to avoid numerical issues
            branch[tdpf_lines, BR_R] = r = r_ref_pu * (1 + alpha_pu * np.clip(np.nan_to_num(T - t_ref_pu), -50, 250 / T_base))
            # todo expansion with SSC and VSC (that are not controllable)
            Ybus, Yf, Yt = makeYbus(baseMVA, bus, branch)
            g, b = calc_g_b(r, x)

        # todo: adjust the SSC J function to take care about the Ybus_ssc_not_controllable instead
        J = create_jacobian_matrix(Ybus+Ybus_ssc_not_controllable+Ybus_vsc_not_controllable, V, ref, refpvpq, pvpq, pq, createJ, pvpq_lookup, nref, npv, npq, numba, slack_weights, dist_slack)

        if tdpf:
            # p.u. values for T, a1, a2, I, S
            # todo: distributed_slack works fine if sn_mva is rather high (e.g. 1000), otherwise no convergence. Why?
            J = create_J_tdpf(branch, tdpf_lines, alpha_pu, r_ref_pu, refpvpq if dist_slack else pvpq, pq, pvpq_lookup,
                              pq_lookup, tau, tdpf_delay_s, Vm, Va, r_theta_pu, J, r, x, g)

        if any_facts_controllable:
            K_J = vstack([eye(J.shape[0], format="csr"),
                          csr_matrix((num_facts_controllable, J.shape[0]))], format="csr")
            J = K_J * J * K_J.T  # this extends the J matrix with 0-rows and 0-columns
        if any_svc:
            J_m_svc = create_J_modification_svc(J, svc_buses, refpvpq if dist_slack else pvpq, pq, pq_lookup, Vm,
                                                x_control_svc, svc_x_l_pu, svc_x_cvar_pu, num_svc_controllable,
                                                svc_controllable)
            J = J + J_m_svc
        if any_tcsc:
            J_m_tcsc = create_J_modification_tcsc(J, V, y_tcsc_pu, x_control_tcsc, tcsc_controllable, tcsc_x_l_pu,
                                                  tcsc_fb, tcsc_tb, refpvpq if dist_slack else pvpq, pq, pvpq_lookup,
                                                  pq_lookup, num_svc_controllable)
            J = J + J_m_tcsc
        if any_ssc_controllable:
            J_m_ssc = create_J_modification_ssc_vsc(J, V, Vm, ssc_y_pu[ssc_controllable], ssc_fb[ssc_controllable],
                                                    ssc_tb[ssc_controllable], refpvpq if dist_slack else pvpq, pq,
                                                    pvpq_lookup, pq_lookup, ssc_mode_ac == 0,
                                                    np.full_like(ssc_mode_ac, False, dtype=bool))
            J = J + J_m_ssc
        if any_vsc_controllable:
            J_m_vsc = create_J_modification_ssc_vsc(J, V, Vm, vsc_y_pu[vsc_controllable], vsc_fb[vsc_controllable],
                                                    vsc_tb[vsc_controllable], refpvpq if dist_slack else pvpq, pq,
                                                    pvpq_lookup, pq_lookup,
                                                    (vsc_mode_ac == VSC_MODE_AC_V) | (vsc_mode_ac == VSC_MODE_AC_SL),
                                                    vsc_mode_ac == VSC_MODE_AC_SL)
            J = J + J_m_vsc
        if any_branch_dc:
            J_m_hvdc = create_J_modification_hvdc(J, V_dc, Ybus_hvdc, Ybus_vsc_dc, vsc_g_pu, vsc_gl_pu, dc_p,
                                                  dc_p_lookup, vsc_dc_fb, vsc_dc_tb, vsc_dc_mode_v, vsc_dc_mode_p)
            J = J + J_m_hvdc

        dx = -1 * spsolve(J, F, permc_spec=permc_spec, use_umfpack=use_umfpack)
        # update voltage
        if dist_slack:
            slack = slack + dx[j0:j1]
        if npv and not iwamoto:
            Va[pv] = Va[pv] + dx[j1:j2]
        if npq and not iwamoto:
            Va[pq] = Va[pq] + dx[j3:j4]
            Vm[pq] = Vm[pq] + dx[j5:j6]
        if any_svc_controllable:
            x_control_svc[svc_controllable] += dx[j6:j6a]
        if any_tcsc_controllable:
            x_control_tcsc[tcsc_controllable] += dx[j6a:j6b]
        if any_branch_dc:
            V_dc[dc_p] += dx[j6d:j6e].real

        if tdpf:
            T = T + dx[j8:][tdpf_lines]

        # iwamoto multiplier to increase convergence
        if iwamoto and not tdpf:
            Vm, Va = _iwamoto_step(Ybus, J, F, dx, pq, npv, npq, dVa, dVm, Vm, Va, pv, j1, j2, j3, j4, j5, j6)

        V = Vm * exp(1j * Va)
        Vm = abs(V)  # update Vm and Va again in case
        Va = angle(V)  # we wrapped around with a negative Vm

        if v_debug:
            Vm_it = column_stack((Vm_it, Vm))
            Va_it = column_stack((Va_it, Va))

        if voltage_depend_loads:
            Sbus = makeSbus(baseMVA, bus, gen, vm=Vm)

        if any_svc_controllable:
            Ybus_svc = makeYbus_svc(Ybus, x_control_svc, svc_x_l_pu, svc_x_cvar_pu, svc_buses)

        if any_tcsc_controllable:
            y_tcsc_pu = -1j * calc_y_svc_pu(x_control_tcsc, tcsc_x_l_pu, tcsc_x_cvar_pu)
            Ybus_tcsc = make_Ybus_facts(tcsc_fb, tcsc_tb, y_tcsc_pu, Ybus.shape[0])
            # Ybus_tcsc = makeYbus_tcsc(Ybus, x_control_tcsc, tcsc_x_l_pu, tcsc_x_cvar_pu, tcsc_fb, tcsc_tb)
        # Ybus_ssc does not change
        # if any_ssc_controllable:
        #     Ybus_ssc = makeYbus_ssc(Ybus, ssc_y_pu, ssc_fb, ssc_tb, any_ssc)

        F = _evaluate_Fx(Ybus + Ybus_svc + Ybus_tcsc + Ybus_ssc + Ybus_vsc, V, Sbus, ref, pv, pq, slack_weights, dist_slack, slack)
        if any_facts_controllable or any_ssc_controllable or any_vsc_controllable:
            mis_facts = _evaluate_Fx_facts(V, pq, svc_buses[svc_controllable], svc_set_vm_pu[svc_controllable],
                                           tcsc_controllable, tcsc_set_p_pu, tcsc_tb, Ybus_tcsc,
                                           ssc_fb[ssc_controllable], ssc_tb[ssc_controllable], Ybus_ssc,
                                           ssc_controllable, ssc_set_vm_pu, F, pq_lookup, vsc_controllable,
                                           vsc_fb[vsc_controllable], vsc_tb[vsc_controllable], Ybus_vsc, Ybus_vsc_dc,
                                           Yf_vsc_dc, Yt_vsc_dc, Yf_vsc, vsc_mode_ac, vsc_mode_dc, vsc_value_ac,
                                           vsc_value_dc, vsc_dc_fb, vsc_dc_tb, vsc_dc_mode_v, vsc_dc_mode_p, vsc_gl_pu,
                                           V_dc, Ybus_hvdc, num_branch_dc, P_dc, dc_p, dc_ref, dc_b2b, dc_p_lookup,
                                           dc_ref_lookup, dc_b2b_lookup, P_dc_sum_sl)
            F = r_[F, mis_facts]

        if tdpf:
            i_square_pu, p_loss_pu = calc_i_square_p_loss(branch, tdpf_lines, g, b, Vm, Va)
            if tdpf_update_r_theta:
                r_theta_pu = calc_r_theta(t_air_pu, a0, a1, a2, i_square_pu, p_loss_pu)
            T_calc = calc_T_frank(p_loss_pu, t_air_pu, r_theta_pu, tdpf_delay_s, T0, tau)
            F_t[tdpf_lines] = T - T_calc
            F = r_[F, F_t]

        converged = _check_for_convergence(F, tol)

    # write q_svc, x_control in ppc["bus"] and then later calculate q_mvar for net.res_shunt
    # todo: move to pf.run_newton_raphson_pf.ppci_to_pfsoln
    if any_svc:
        y_svc_pu = calc_y_svc_pu(x_control_svc, svc_x_l_pu, svc_x_cvar_pu)
        q_svc_pu = np.square(np.abs(V[svc_buses])) * y_svc_pu
        svc[svc_idx, SVC_Q] = q_svc_pu * baseMVA
        svc[svc_idx, SVC_THYRISTOR_FIRING_ANGLE] = x_control_svc
        svc[svc_idx, SVC_X_PU] = 1 / y_svc_pu

    # todo: move to pf.run_newton_raphson_pf.ppci_to_pfsoln
    if any_tcsc:
        Yf_tcsc, Yt_tcsc = makeYft_tcsc(Ybus_tcsc, tcsc_fb, tcsc_tb, x_control_tcsc, tcsc_x_l_pu, tcsc_x_cvar_pu)
        # todo use make_Ybus_facts, make_Yft_facts
        baseI = baseMVA / (bus[tcsc_tb, BASE_KV] * sqrt(3))
        i_tcsc_f = Yf_tcsc.dot(V)
        i_tcsc_t = Yt_tcsc.dot(V)
        s_tcsc_f = np.conj(i_tcsc_f) * V[tcsc_fb] * baseMVA
        s_tcsc_t = np.conj(i_tcsc_t) * V[tcsc_tb] * baseMVA
        tcsc[tcsc_branches, TCSC_THYRISTOR_FIRING_ANGLE] = x_control_tcsc
        tcsc[tcsc_branches, TCSC_PF] = s_tcsc_f.real
        tcsc[tcsc_branches, TCSC_QF] = s_tcsc_f.imag
        tcsc[tcsc_branches, TCSC_PT] = s_tcsc_t.real
        tcsc[tcsc_branches, TCSC_QT] = s_tcsc_t.imag
        tcsc[tcsc_branches, TCSC_IF] = np.abs(i_tcsc_f) * baseI
        tcsc[tcsc_branches, TCSC_IT] = np.abs(i_tcsc_t) * baseI
        tcsc[tcsc_branches, TCSC_X_PU] = 1 / calc_y_svc_pu(x_control_tcsc, tcsc_x_l_pu, tcsc_x_cvar_pu)

    # todo: move to pf.run_newton_raphson_pf.ppci_to_pfsoln
    if any_ssc:
        # todo use make_Ybus_facts, make_Yft_facts
        Yf_ssc, Yt_ssc = make_Yft_facts(ssc_fb, ssc_tb, ssc_y_pu, Ybus_ssc.shape[0])
        s_ssc_f = conj(Yf_ssc.dot(V)) * V[ssc_fb] * baseMVA
        s_ssc_t = conj(Yt_ssc.dot(V)) * V[ssc_tb] * baseMVA
        ssc[ssc_branches, SSC_Q] = s_ssc_f.imag

    # todo move this to pfsoln
    if any_vsc:
        # Yf_vsc, Yt_vsc = make_Yft_facts(vsc_fb, vsc_tb, vsc_y_pu, Ybus_vsc.shape[0])
        s_vsc_f = conj(Yf_vsc.dot(V)) * V[vsc_fb] * baseMVA
        s_vsc_t = conj(Yt_vsc.dot(V)) * V[vsc_tb] * baseMVA
        vsc[vsc_branches, VSC_P] = s_vsc_f.real
        vsc[vsc_branches, VSC_Q] = s_vsc_f.imag

        # Yf_vsc_dc, Yt_vsc_dc = make_Yft_facts(vsc_dc_fb, vsc_dc_tb, vsc_g_pu, num_bus_dc)
        i_vsc_dc_f = Yf_vsc_dc.dot(V_dc)
        p_vsc_dc_f = i_vsc_dc_f * V_dc[vsc_dc_fb] * baseMVA
        i_vsc_dc_t = Yt_vsc_dc.dot(V_dc)
        p_vsc_dc_t = i_vsc_dc_t * V_dc[vsc_dc_tb] * baseMVA

        vsc[vsc_branches, VSC_P_DC] = p_vsc_dc_f
        # print(p_vsc_dc_f, p_vsc_dc_t)
        # print(s_vsc_f.real, s_vsc_t.real)
        # no_load_losses = vsc_gl_pu * np.square(V_dc[vsc_dc_fb])
        # print(f"{no_load_losses=}")
        # print(f"{p_vsc_dc_t + p_vsc_dc_f}")
        # print(f"{p_vsc_dc_t + p_vsc_dc_f + no_load_losses}")

    if len(relevant_bus_dc) > 0:
        # duplication with "if any branch dc" because it is possible to have only vsc and no dc lines:
        bus_dc[relevant_bus_dc, DC_VM] = V_dc
        # bus_dc[relevant_bus_dc, DC_PD] = -P_dc

    # todo move this to pfsoln
    if any_branch_dc:
        Yf_hvdc, Yt_hvdc = make_Yft_facts(hvdc_fb, hvdc_tb, hvdc_y_pu, num_bus_dc)
        # Pbus_dc = V_dc * (Ybus_hvdc + Ybus_vsc_dc).dot(V_dc)
        Pbus_dc = V_dc * Ybus_hvdc.dot(V_dc)
        bus_dc[relevant_bus_dc, DC_PD] = Pbus_dc
        i_hvdc_f = Yf_hvdc.dot(V_dc)
        p_hvdc_f = i_hvdc_f * V_dc[hvdc_fb] * baseMVA
        i_hvdc_t = Yt_hvdc.dot(V_dc)
        p_hvdc_t = i_hvdc_t * V_dc[hvdc_tb] * baseMVA
        branch_dc[hvdc_branches, DC_PF] = p_hvdc_f.real
        branch_dc[hvdc_branches, DC_IF] = i_hvdc_f.real
        branch_dc[hvdc_branches, DC_PT] = p_hvdc_t.real
        branch_dc[hvdc_branches, DC_IT] = i_hvdc_t.real

    # todo: remove this
    # Ybus, Yf, Yt = makeYbus(baseMVA, bus, branch)
    # Sf =  conj(Yf.dot(V)) * V[[0, 0]] * baseMVA
    # St =  conj(Yt.dot(V)) * V[[1, 2]] * baseMVA
    # print("P from:", Sf.real)
    # print("P to:", St.real)
    # print("Sbus:", Sbus.real)

    # because we now have updates of the Ybus matrices due to TDPF, SVC, TCSC, SSC,
    # we are interested in storing them for later use:
    ppci["internal"]["Ybus"] = Ybus
    ppci["internal"]["Ybus_svc"] = Ybus_svc
    ppci["internal"]["Ybus_tcsc"] = Ybus_tcsc
    ppci["internal"]["Ybus_ssc"] = Ybus_ssc
    ppci["internal"]["Ybus_vsc"] = Ybus_vsc
    ppci["internal"]["Ybus_hvdc"] = Ybus_hvdc
    # ppci["internal"]["Yf_tcsc"] = Yf_tcsc
    # ppci["internal"]["Yt_tcsc"] = Yt_tcsc
    ppci["internal"]["tcsc_fb"] = tcsc_fb
    ppci["internal"]["tcsc_tb"] = tcsc_tb

    return V, converged, i, J, Vm_it, Va_it, r_theta_pu / baseMVA * T_base, T * T_base


def _evaluate_Fx(Ybus, V, Sbus, ref, pv, pq, slack_weights=None, dist_slack=False, slack=None):
    # evalute F(x)
    if dist_slack:
        # we include the slack power (slack * contribution factors) in the mismatch calculation
        mis = V * conj(Ybus * V) - Sbus + slack_weights * slack
        F = r_[mis[ref].real, mis[pv].real, mis[pq].real, mis[pq].imag]
    else:
        mis = V * conj(Ybus * V) - Sbus
        F = r_[mis[pv].real, mis[pq].real, mis[pq].imag]
    return F


def _evaluate_Fx_facts(V,pq ,svc_buses=None, svc_set_vm_pu=None, tcsc_controllable=None, tcsc_set_p_pu=None, tcsc_tb=None,
                       Ybus_tcsc=None, ssc_fb=None, ssc_tb=None, Ybus_ssc=None,
                       ssc_controllable=None, ssc_set_vm_pu=None, old_F=None, pq_lookup=None,
                       vsc_controllable=None, vsc_fb=None, vsc_tb=None,
                       Ybus_vsc=None, Ybus_vsc_dc=None, Yf_vsc_dc=None, Yt_vsc_dc=None, Yf_vsc=None,
                       vsc_mode_ac=None, vsc_mode_dc=None, vsc_value_ac=None, vsc_value_dc=None, vsc_dc_fb=None, vsc_dc_tb=None,
                       vsc_dc_mode_v=None, vsc_dc_mode_p=None, vsc_gl_pu=None,
                       V_dc=None, Ybus_hvdc=None, num_branch_dc=None, P_dc=None, dc_p=None, dc_ref=None, dc_b2b=None,
                       dc_p_lookup=None, dc_ref_lookup=None, dc_b2b_lookup=None, P_dc_sum_sl=None):
    mis_facts = np.array([], dtype=np.float64)

    if svc_buses is not None and len(svc_buses) > 0:
        mis_facts = np.r_[mis_facts, np.abs(V[svc_buses]) - svc_set_vm_pu]

    if tcsc_tb is not None and len(tcsc_tb) > 0:
        if np.any(tcsc_controllable):
            Sbus_tcsc = V * conj(Ybus_tcsc * V)
            mis_tcsc = Sbus_tcsc[tcsc_tb[tcsc_controllable]].real - tcsc_set_p_pu
            mis_facts = np.r_[mis_facts, mis_tcsc]

    # if ssc_fb is not None and len(ssc_fb) > 0:
    if np.any(ssc_controllable):
        mis_ssc_v = np.abs(V[ssc_fb]) - ssc_set_vm_pu
        # mis_facts = np.r_[mis_facts, mis_ssc_p, mis_ssc_v]
        # old_F[-(len(pq)+len(ssc_fb))] = mis_ssc_p
        # old_F[-len(ssc_fb)] = mis_ssc_v
        old_F[-len(pq)+pq_lookup[ssc_tb]] = mis_ssc_v

    # todo: fix this condition (because there can be a mix of dc_p and dc_ref buses) - use np.where?
    if num_branch_dc > 0:  # todo if hvdc lines
        # This part only relevant for DC lines:
        Pbus_hvdc = V_dc * Ybus_hvdc.dot(V_dc)
        Pbus_vsc = V_dc * Ybus_vsc_dc.dot(V_dc)
        p_vsc_dc_f = V_dc[vsc_dc_fb] * Yf_vsc_dc.dot(V_dc)
        p_vsc_dc_t = V_dc[vsc_dc_tb] * Yt_vsc_dc.dot(V_dc)
        # vsc_dc_ref = vsc_mode_dc == VSC_MODE_DC_V
        # if np.any(vsc_dc_ref):
        #     P_dc[vsc_dc_tb[vsc_dc_ref]] = Pbus_hvdc[vsc_dc_fb[vsc_dc_ref]]
        # Pbus_hvdc = V_dc * (Ybus_hvdc+Ybus_vsc_dc).dot(V_dc)

        mis_hvdc = Pbus_hvdc - P_dc  # todo vsc
        # mis_hvdc = Pbus_hvdc - np.zeros_like(Pbus_hvdc)
        # mis_hvdc = np.zeros_like(Pbus_hvdc)
        if np.any(vsc_dc_mode_p):
            mis_hvdc[vsc_dc_tb[vsc_dc_mode_p]] = p_vsc_dc_f[vsc_dc_mode_p] - vsc_value_dc[vsc_dc_mode_p]

        if np.any(vsc_dc_mode_v):
            mis_hvdc[vsc_dc_tb[vsc_dc_mode_v]] = V_dc[vsc_dc_fb[vsc_dc_mode_v]] - vsc_value_dc[vsc_dc_mode_v]

        mis_facts = np.r_[mis_facts, mis_hvdc[dc_p]]
    else:
        Pbus_hvdc = -P_dc

    if np.any(vsc_controllable):
        Sbus_vsc = V * conj(Ybus_vsc * V)
        sf_vsc = V[vsc_fb] * conj(Yf_vsc.dot(V))
        ###### Mismatch for the first VSC variable:
        # the mismatch value is written for the aux bus, but calculated based on the AC bus
        ac_mode_v = (vsc_mode_ac == VSC_MODE_AC_V) | (vsc_mode_ac == VSC_MODE_AC_SL)
        if np.any(ac_mode_v):
            mis_vsc_v = np.abs(V[vsc_fb[ac_mode_v]]) - vsc_value_ac[ac_mode_v]
            old_F[-len(pq)+pq_lookup[vsc_tb[ac_mode_v]]] = mis_vsc_v

        # Mismatch for Q:
        ac_mode_q = vsc_mode_ac == VSC_MODE_AC_Q
        if np.any(ac_mode_q):
            # unique_vsc_q_bus, c_q, _ = _sum_by_group(vsc_fb[ac_mode_q], np.ones_like(vsc_fb[ac_mode_q]), np.ones_like(vsc_fb[ac_mode_q]))
            # count_q = np.zeros(vsc_fb.max() + 1, dtype=np.int64)
            # count_q[unique_vsc_q_bus] = c_q
            # mis_vsc_q = Sbus_vsc[vsc_fb[ac_mode_q]].imag / count_q[vsc_fb[ac_mode_q]] - vsc_value_ac[ac_mode_q]  # todo test for when they share same bus
            mis_vsc_q = sf_vsc.imag[ac_mode_q] - vsc_value_ac[ac_mode_q]  # todo test for when they share same bus
            old_F[-len(pq) + pq_lookup[vsc_tb[ac_mode_q]]] = mis_vsc_q


        ##### Mismatch for the second VSC variable:
        # Mismatch for AC slack - delta
        ac_mode_sl = vsc_mode_ac == VSC_MODE_AC_SL
        if np.any(ac_mode_sl):
            mis_vsc_delta = np.angle(V[vsc_fb[ac_mode_sl]]) - 0  # <- here we set delta set point to zero, but can be a parameter in the future
            old_F[-len(pq) * 2 + pq_lookup[vsc_tb[ac_mode_sl]]] = mis_vsc_delta
            # this connects the AC slack result and the DC bus P set-point:
            vsc_slack_p = -Sbus_vsc[vsc_tb[ac_mode_sl]].real
            vsc_slack_p_dc_bus, vsc_slack_p_dc, _ = _sum_by_group(vsc_dc_tb[ac_mode_sl], vsc_slack_p, vsc_slack_p)
            P_dc[vsc_slack_p_dc_bus] = P_dc_sum_sl + vsc_slack_p_dc

        # find the connection between the DC buses and VSC buses
        # find the slack DC buses
        # find the P DC buses
        vsc_set_p_pu = np.zeros(len(vsc_fb[vsc_controllable]), dtype=np.float64)  # power sign convention issue
        # vsc_dc_p_bus = vsc_dc_fb[vsc_controllable][np.isin(vsc_dc_fb[vsc_controllable], dc_p)]
        # vsc_dc_ref_bus = vsc_dc_fb[vsc_controllable][np.isin(vsc_dc_fb[vsc_controllable], dc_ref)]
        # vsc_dc_b2b_bus = vsc_dc_fb[vsc_controllable][np.isin(vsc_dc_fb[vsc_controllable], dc_b2b)]

        vsc_ac_sl = vsc_mode_ac == VSC_MODE_AC_SL
        vsc_dc_p = (vsc_mode_dc == VSC_MODE_DC_P) & (vsc_mode_ac != VSC_MODE_AC_SL)
        vsc_dc_ref = vsc_mode_dc == VSC_MODE_DC_V
        # vsc_b2b_p = np.isin(vsc_dc_fb, vsc_dc_b2b_bus) & (vsc_mode_dc == VSC_MODE_DC_P)
        # vsc_b2b_ref = np.isin(vsc_dc_fb, vsc_dc_b2b_bus) & (vsc_mode_dc == VSC_MODE_DC_V)

        # unique_vsc_dc_bus, c_ref, c_b2b_ref = _sum_by_group(vsc_dc_fb, vsc_dc_ref.astype(np.float64), vsc_b2b_ref.astype(np.float64))
        # count_ref = np.zeros(vsc_dc_fb.max() + 1, dtype=np.int64)
        # count_b2b_ref = np.zeros(vsc_dc_fb.max() + 1, dtype=np.int64)
        # count_ref[unique_vsc_dc_bus] = c_ref
        # count_b2b_ref[unique_vsc_dc_bus] = c_b2b_ref

        # unique_vsc_p_bus, c_p, _ = _sum_by_group(vsc_fb[vsc_dc_p], np.ones_like(vsc_fb[vsc_dc_p]), np.ones_like(vsc_fb[vsc_dc_p]))
        # count_p = np.zeros(vsc_fb.max() + 1, dtype=np.int64)
        # count_p[unique_vsc_p_bus] = c_p


        # todo fix this part
        # a configuration of B2B bus connecting also with DC lines could be implemented here in the future if needed:
        # if np.any(vsc_ac_sl):
        #     P_dc[vsc_dc_tb[vsc_ac_sl]] = -Sbus_vsc[vsc_tb[vsc_controllable]].real[vsc_ac_sl]
        if np.any(vsc_dc_p):
            # vsc_set_p_pu[vsc_dc_p] = -P_dc[dc_p][dc_p_lookup[vsc_dc_p_bus]]
            # vsc_set_p_pu[vsc_dc_p] = vsc_value_dc[vsc_dc_p] * count_p[vsc_fb[vsc_dc_p]] # todo test for when they share same bus
            # vsc_set_p_pu[vsc_dc_p] = vsc_value_dc[vsc_dc_p]  # todo consider count
            no_load_losses = vsc_gl_pu * np.square(V_dc[vsc_dc_fb])
            vsc_set_p_pu[vsc_dc_p] = -p_vsc_dc_t[vsc_dc_p] #+ no_load_losses[vsc_dc_p]
            # P_dc[dc_p] = -Sbus_vsc[vsc_tb[vsc_controllable]].real[vsc_dc_p]
        if np.any(vsc_dc_ref):
            # vsc_set_p_pu[vsc_dc_ref] = -Pbus_hvdc[dc_ref][dc_ref_lookup[vsc_dc_ref_bus]] / count_ref[vsc_dc_bus[vsc_dc_ref]]
            vsc_set_p_pu[vsc_dc_ref] = -Pbus_hvdc[vsc_dc_tb[vsc_dc_ref]] #- Pbus_hvdc[vsc_dc_fb[vsc_dc_ref]]
            # vsc_set_p_pu[vsc_dc_ref] = -P_dc[vsc_dc_tb[vsc_dc_ref]]
            # # vsc_set_p_pu[vsc_mode_dc == 1] = -P_dc[dc_ref][dc_ref_lookup[vsc_dc_ref_bus]]  # dc_p
            #
            # # Create a list of tuples containing the original value and its index
            # vsc_dc_bus_idx = np.array(list(enumerate(vsc_dc_bus)))[vsc_dc_ref]  ##TODO: consider define the vscs indices list instead
            #
            # # Sort the list of tuples based on the values
            # vsc_dc_bus_sorted = sorted(vsc_dc_bus_idx, key=lambda x: x[1])
            #
            # # Create the output array by extracting the indices from the sorted list of tuples
            # vsc_dc_bus_sorted_idx = [i for i, value in vsc_dc_bus_sorted]
            #
            # vsc_set_p_pu[vsc_dc_bus_sorted_idx] = -Pbus_hvdc[dc_ref][dc_ref_lookup[vsc_dc_ref_bus]] / count_ref[vsc_dc_bus[vsc_dc_ref]]

        # if len(dc_b2b):
        #     # vsc_set_p_pu[vsc_mode_dc == 1] = -P_dc[dc_ref][dc_ref_lookup[vsc_dc_ref_bus]]  # dc_p
        #     vsc_set_p_pu[vsc_b2b_p] = vsc_value_dc[vsc_b2b_p]
        #     # vsc_set_p_pu[vsc_b2b_ref] = P_dc[dc_b2b][dc_b2b_lookup[vsc_dc_b2b_bus[vsc_b2b_ref]]] / count_b2b_ref[vsc_dc_bus[vsc_b2b_ref]]
        #     vsc_set_p_pu[vsc_b2b_ref] = P_dc[dc_b2b][dc_b2b_lookup[vsc_dc_fb[vsc_b2b_ref]]] / count_b2b_ref[vsc_dc_fb[vsc_b2b_ref]]
        ####  here used vsc_tb refereing to the q bus
        # S_temp = Sbus_vsc.real
        # S_temp[vsc_tb[vsc_dc_p]] /= count_p[vsc_fb[vsc_dc_p]]
        # mis_vsc_p = S_temp[vsc_tb[vsc_controllable]] - vsc_set_p_pu  # this is coupling the AC and the DC sides
        mis_vsc_p = Sbus_vsc[vsc_tb[vsc_controllable]].real - vsc_set_p_pu  # this is coupling the AC and the DC sides # todo old
        # todo: adjust the lookup to work with 1) VSC at ext_grid bus 2) only 1 VSC connected to HVDC line
        old_F[-len(pq) * 2 + pq_lookup[vsc_tb[~ac_mode_sl]]] = mis_vsc_p[~ac_mode_sl]

    return mis_facts


def _check_for_convergence(F, tol):
    # calc infinity norm
    return linalg.norm(F, inf) < tol




# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

"""
Defines constants for named column indices to VSC matrix.

Some examples of usage, after defining the constants using the lines above,
are::

    vsc[3, VSC_STATUS] = 0                  # take VSC 4 out of service
    VSC_Q = vsc[:, VSC_Q]                   # extract reactive power vector

The index, name, and meaning of each column of the VSC matrix are given below:

columns 0-12 must be included in the input matrix (vsc in ppc dict)
    0.  C{VSC_BUS}                      'from' AC bus number
    1.  C{VSC_INTERNAL_BUS}             Internal bus number for VSC
    2.  C{VSC_INTERNAL_BUS_DC}          Internal DC bus number for VSC
    3.  C{VSC_BUS_DC}                   'to' DC bus number
    4.  C{VSC_R}                        VSC resistance (p.u.)
    5.  C{VSC_X}                        VSC reactance (p.u.)
    6.  C{VSC_R_DC}                     VSC dc-side resistance (p.u.)
    7.  C{VSC_MODE_AC}                  AC mode selector, 0 - vm_pu, 1 - q_mvar
    8.  C{VSC_VALUE_AC}                 AC mode value (p.u.)
    9.  C{VSC_MODE_DC}                  DC mode selector, 0 - vm_pu, 1 - p_mw
    10  C{VSC_VALUE_DC}                 DC mode value (p.u.)
    11. C{VSC_STATUS}                   Initial VSC status, 1 - in service, 0 - out of service
    12. C{VSC_CONTROLLABLE}             Controllability flag for VSC
    13. C{VSC_P}                        AC bus resultant active power (MW)
    14. C{VSC_Q}                        AC bus resultant reactive power (MVAr)
    15. C{VSC_P_DC}                     DC bus resultant active power (MW)

"""

VSC_MODE_AC_V = 0
VSC_MODE_AC_Q = 1
VSC_MODE_AC_SL = 2
VSC_MODE_DC_V = 0
VSC_MODE_DC_P = 1

# define the indices
VSC_BUS = 0  # f, from bus number
VSC_INTERNAL_BUS = 1
VSC_INTERNAL_BUS_DC = 2
VSC_BUS_DC = 3  # f, from bus number
VSC_R = 4  # (p.u.)
VSC_X = 5  # (p.u.)
VSC_R_DC = 6  # (p.u.)
VSC_PL_DC = 7  # p.u., specifies the no-load losses (initially input as MW and converted to p.u. for ppc)
VSC_MODE_AC = 8  # 0 - vm_pu, 1 - q_mvar, 2 - slack
VSC_VALUE_AC = 9
VSC_MODE_DC = 10  # 0 - vm_pu, 1 - p_mw
VSC_VALUE_DC = 11
VSC_STATUS = 12  # initial branch status, 1 - in service, 0 - out of service
VSC_CONTROLLABLE = 13

# for results:
VSC_P = 14  # result for P
VSC_Q = 15  # result for Q
VSC_P_DC = 16  # result for P


vsc_cols = 17



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np

from pandapower.pypower.idx_bus import PD, QD
from pandapower.pf.ppci_variables import _get_pf_variables_from_ppci
from pandapower.pf.pfsoln_numba import pfsoln
from pandapower.results import _copy_results_ppci_to_ppc, _extract_results_se, init_results
from pandapower.auxiliary import get_values


def _calc_power_flow(ppci, V):
    # store results for all elements
    # calculate branch results (in ppc_i)
    baseMVA, bus, gen, branch, svc, tcsc, ssc, vsc, ref, pv, pq, *_, ref_gens = _get_pf_variables_from_ppci(ppci)
    Ybus, Yf, Yt = ppci['internal']['Ybus'], ppci['internal']['Yf'], ppci['internal']['Yt']
    ppci['bus'], ppci['gen'], ppci['branch'] = \
        pfsoln(baseMVA, bus, gen, branch, svc, tcsc, ssc, vsc, Ybus, Yf, Yt, V, ref, ref_gens)

    # calculate bus power injections
    Sbus = np.multiply(V, np.conj(Ybus * V)) * baseMVA
    ppci["bus"][:, PD] = -Sbus.real  # saved in MW, injection -> demand
    ppci["bus"][:, QD] = -Sbus.imag  # saved in Mvar, injection -> demand
    return ppci


def _extract_result_ppci_to_pp(net, ppc, ppci):
    # convert to pandapower indices
    ppc = _copy_results_ppci_to_ppc(ppci, ppc, mode="se")

    # inits empty result tables
    init_results(net, mode="se")

    # writes res_bus.vm_pu / va_degree and branch res
    _extract_results_se(net, ppc)

    # additionally, write bus power demand results (these are not written in _extract_results)
    mapping_table = net["_pd2ppc_lookups"]["bus"]
    net.res_bus_est.index = net.bus.index
    net.res_bus_est.p_mw = get_values(ppc["bus"][:, 2], net.bus.index.values,
                                      mapping_table)
    net.res_bus_est.q_mvar = get_values(ppc["bus"][:, 3], net.bus.index.values,
                                        mapping_table)
    # overwrite power values for buses that were merged because they would not have the same power inj
    # as the bus they were merged to
    merged_bus = net["_pd2ppc_lookups"]["merged_bus"]
    merged_bus_idx = np.where(merged_bus == True)[0]
    net.res_bus_est.loc[merged_bus_idx, 'p_mw'] = 0
    net.res_bus_est.loc[merged_bus_idx, "q_mvar"] = 0
    # add shunt power because the injection at the node computed via Ybus is only the extra injection on top of the shunt
    for element in ["shunt", "ward", "xward"]:
        if ~net[element].empty:
            for i in range(net[element].shape[0]):
                bus = net[element].bus.iloc[i]
                if element == "shunt":
                    Sn = complex(net[element].p_mw.iloc[i],net[element].q_mvar.iloc[i])*net[element].step.iloc[i]
                    Ysh = Sn / (net[element].vn_kv.iloc[i]**2)
                else:
                    Sn = complex(net[element].pz_mw.iloc[i],net[element].qz_mvar.iloc[i])
                    Ysh = Sn / (net.bus.loc[bus,"vn_kv"]**2)
                V = net["res_bus_est"].loc[bus,"vm_pu"]*net["bus"].loc[bus,"vn_kv"]
                Sinj = Ysh*(V**2)
                net["res_bus_est"].loc[bus,"p_mw"] += Sinj.real
                net["res_bus_est"].loc[bus,"q_mvar"] += Sinj.imag
                if element == "shunt":
                    element_res_est = "res_" + element + "_est"
                    net[element_res_est].loc[net[element].loc[:,"bus"]==bus,"p_mw"] = Sinj.real
                    net[element_res_est].loc[net[element].loc[:,"bus"]==bus,"q_mvar"] = Sinj.imag
                    net[element_res_est].loc[net[element].loc[:,"bus"]==bus,"vm_pu"] = net["res_bus_est"].loc[bus,"vm_pu"]
    return net


def eppci2pp(net, ppc, eppci):
    # calculate the branch power flow and bus power injection based on the estimated voltage vector
    eppci = _calc_power_flow(eppci, eppci.V)

    # extract the result from ppci to ppc and pandpower network
    net = _extract_result_ppci_to_pp(net, ppc, eppci)
    return net
